#ifndef X86INTER_CSET_H_
#define X86INTER_CSET_H_

/* Define asm shortcuts for typical operations. We will need only a
 * handful of such primitives.
 *
 * Basically, we want to do a 32x32 -> 64 multiplication. The mull
 * operation does this. We have several possible flavors:
 *
 * m_put_put(a,b,lo,hi) : put low and high_word to args 3 and 4
 * m_add(a,b,lo) : put low word to arg. 3, discard high part of product.
 * 
 * When using a dataword in : 
 *
 * mc_put_put(a,b,c,lo,hi) : puts a * b + c to [lo:hi] (no carry out)
 * mc_put(a,b,c,lo) : same, but discarding high part.
 *
 * We also have a streamed version of these things that does the
 * following:
 *
* x__I_sm(a,b,r) : add cx to r, put (a * b + bx + carry out) to [bx:cx]
 * 		  note that then bx is the low part, and cx is the high
 * 		  part ; hence we have the dual:
 * x_I__sm(a,b,r).
 *
 * And, to finish:
 *
 * x__I_fm(a,b,lo,hi) : add cx to lo, add (a * b + bx + carry out) to hi
 * x_I__fm(a,b,lo,hi) : add bx to lo, add (a * b + cx + carry out) to hi
 *
 * The extraneous underscores serve as a hint to notice quickly that the
 * swap between [bx:cx] and [cx:bx] is correctly done.
 */
/* put low and high word */

#define m_put_put(a,b,lo,hi)			\
		"movl	" a ", %%eax	\n"	\
		"mull	" b "		\n"	\
		"movl	%%eax, " lo "	\n"	\
		"movl	%%edx, " hi "	\n"

/* add low word, discard high */
#define	m_add(s1,s2,r)				\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	%%eax, " r "	\n"

/* puts s1 * s2 + c to r1,r2 */
/* r2,r3 can then be used as c,r1 in a following replication */
#define mc_put_put(s1,s2,c,r1,r2)		\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	" c ", %%eax	\n"	\
		"adcl	$0, %%edx		\n"	\
		"movl	%%eax, " r1 "	\n"	\
		"movl	%%edx, " r2 "	\n"

		/* better ? */
#define xxxmc_put_put(s1,s2,c,r1,r2)		\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	" c ", %%eax	\n"	\
		"movl	$0, " c "		\n"	\
		"adcl	" c ", %%edx	\n"	\
		"movl	%%eax, " r1 "	\n"	\
		"movl	%%edx, " r2 "	\n"
		
#define mc_put(s1,s2,c,r1)			\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	" c ", %%eax	\n"	\
		"movl	%%eax, " r1 "	\n"

/* Adds c to r1, and puts s1 * s2 + carry out + r2 to r2,r3 */
/* r2,r3 can then be used as c,r2 in a following replication */

/*
 * r1+=c;
 * carry=(r1<c);
 * umul_ppmm(hi,lo,s1,s2);
 * add_carry(carry,r2  ,  lo,carry);
 * r3+=carry;

		s1	s2	c
		  \    /      /
		   mull    /
		    \__ /
		   __/ \_____
		  +     +   |
	         r1---+r2--+r3
 *
 * The sad thing about this diagram is that it proves somehow that seven
 * registers are necessary...
 */
		 
#define MUL_RA_AC_P(s1,s2,c,r1,r2,r3)		\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	" c ", " r1 "	\n"	\
		"adcl	%%eax, " r2 "	\n"	\
		"movl	%%edx, " r3 "	\n"	\
		"adcl	$0, " r3 "	\n"

#define	x_I__sm(a,b,r)	MUL_RA_AC_P(a,b,rB,r,rC,rB)
#define	x__I_sm(a,b,r)	MUL_RA_AC_P(a,b,rC,r,rB,rC)

/* Adds c to r1, and adds s1 * s2 + carry out + y to r2 */
#define MUL_RA_AC_A_MV(s1,s2,c,y,r1,r2)		\
		"movl	" s1 ", %%eax	\n"	\
		"mull	" s2 "		\n"	\
		"addl	" c ", " r1 "	\n"	\
		"adcl	%%eax, " y "	\n"	\
		"addl	" y ", " r2 "	\n"
		
#define	x_I__fm(a,b,lo,hi)	MUL_RA_AC_A_MV(a,b,rB,rC,lo,hi)
#define	x__I_fm(a,b,lo,hi)	MUL_RA_AC_A_MV(a,b,rC,rB,lo,hi)

#define	RESTORE_EBP "popl %%ebp     \n"

#define	SAVE_EBP_AND_SET		\
	"movl	%[r], %%eax	\n"	\
	"pushl	%%ebp		\n"	\
	"movl	%%eax, %%ebp	\n"

#endif	/* X86INTER_CSET_H_ */
