/* $Id: z2n.c 197 2004-10-18 08:37:20Z /CN=thome/emailAddress=thome@loria.fr $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define HAVE_MEMSET 1

#include <gmp.h>

#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "gmp_hacks.h"
#include "mploc.h"
#include "z2.h"
#include "z2x.h"
#include "z2n.h"

#include "ff/rf_layer_gf2n_ntl.hpp"

#include "manu.h"

/* An element of z2n is just a polynomial.  */

struct mplf_private {
	mploc_z2x_t	p;
	struct z2x_reducer_t *	reducer;
};

/* mplf_z2n_create -- Create a Z2N field
 *
 * INPUT: The definition polynomial
 * OUTPUT: The field structure
 */
mplf_t mplf_z2n_create(mploc_z2x_srcptr p)
{
	struct mplf_struct * f;

	f=malloc(sizeof(struct mplf_struct));
#ifdef DEBUG_MPLOC
	f->people=0;
#endif
	f->degree=p->deg;
	f->ramified=0;	/* FIXME : Obviously wrong. Perhaps this
			   could be removed, since we don't care */
	f->priv=malloc(sizeof(struct mplf_private));
	mploc_z2x_init(f->priv->p,p->coeff_field);
	mploc_z2x_set(f->priv->p,p);
	/* Default (dumb) reduction routine via Euclidean division does
	 * handle overlaps. */
	f->priv->reducer=NULL;
	mpz_init_set_ui(f->p,2);

	ATTACH_RESIDUE_FIELD(f);

	return f;
}

mploc_z2x_srcptr mplf_z2n_def_poly(mplf_t f)
{
	return f->priv->p;
}

/* mplf_z2n_destroy -- Frees the data asssociated with a Z2N field.
 *
 * INPUT: A pointer to the field structure.
 * OUTPUT: none ; pointer is set to NULL.
 *
 * All elements attached to the field are assumed to have been cleared
 * previously.
 */
void mplf_z2n_destroy(mplf_t * pf)
{
	mplf_z2n_set_reducer(*pf,NULL);

#ifdef DEBUG_MPLOC
	ASSERT(!(*pf)->people); 
#endif

	DETACH_RESIDUE_FIELD((*pf));

	mploc_z2x_clear((*pf)->priv->p);
	mpz_clear((*pf)->p);
	free((*pf)->priv);
	free(*pf);
	*pf=NULL;
}

/* Do the X^2n / f stuff */
void mplf_z2n_set_reducer(mplf_t f, struct z2x_reducer_t * rf)
{
	if (f->priv->reducer) {
		if (f->priv->reducer->unhook) {
			f->priv->reducer->unhook(f);
			ASSERT(f->priv->reducer==NULL);
		} else {
			/* Since this data is known only via a generic
			 * pointer, I can't know how to copy it. Hence
			 * freeing it is out of by business as well.
			 */
			ASSERT(f->priv->reducer->priv==NULL);
			free(f->priv->reducer);
			f->priv->reducer=NULL;
		}
	}
	/* NULL means we return back to the default reducer */
	if (rf==NULL) return;

	/* This would be good practice. Then I don't have to enforce
	 * policies as to whether structs are ptr-linked or copied
	 */
	ASSERT((rf->hook != NULL) == (rf->unhook != NULL));
	ASSERT((rf->hook != NULL) == (rf->priv != NULL));

	if (rf->hook) {
		rf->hook(f, rf);
	} else {
		f->priv->reducer=malloc(sizeof(struct z2x_reducer_t));
		memcpy(f->priv->reducer,rf,sizeof(struct z2x_reducer_t));
	}
}

/* preinv reducer stuff ; This could possibly go into another file */

#define	MPLF_Z2N__PREINV_REDUCER_MAGIC	0xABCD0001

struct mplf_z2n__preinv_reducer_priv_t {
	mploc_z2x_t preinv;
	mploc_z2x_t svoboda;
	int	autofree;
};

#define	PREINV_REDUCER_PRIV(rf)	((struct mplf_z2n__preinv_reducer_priv_t *)((rf)->priv))
#define	FIELD_PREINV(f)	(PREINV_REDUCER_PRIV((f)->priv->reducer)->preinv)
#define	FIELD_SVOBODA(f)	(PREINV_REDUCER_PRIV((f)->priv->reducer)->svoboda)

static struct z2x_reducer_t * mplf_z2n__preinv_reducer_new(mplf_t f);
static void mplf_z2n__preinv_reducer_delete(struct z2x_reducer_t *);
static void mplf_z2n__preinv_reducer_reduce(mploc_z2x_ptr, mploc_z2x_srcptr);
static void mplf_z2n__preinv_reducer_hook(mplf_t f, struct z2x_reducer_t * rf);
static void mplf_z2n__preinv_reducer_unhook(mplf_t f);

static void mplf_z2n__preinv_reducer_reduce(mploc_z2n_ptr r, mploc_z2n_srcptr a)
{
	mploc_z2x_t a1,q,t,src;
	int n = a->f->degree;

	mploc_z2x_init(a1,a->coeff_field);
	mploc_z2x_init(q,a->coeff_field);
	mploc_z2x_init(t,a->coeff_field);
	mploc_z2x_init(src,a->coeff_field);

	mploc_z2x_set(src,a);

	/* Note that our primary objective is the 2n->n reduction ; this
	 * is the reason why we have precomputed the degree n preinverse.
	 * This implies that the 2n->n reduction has cost 2M(n).
	 *
	 * It is unwise to precompute a larger preinverse for a (1+k)n->n
	 * reduction.
	 *
	 * For k<=1, the following code does the same as working with the
	 * tight preinverse (because of early degree chopping).
	 *
	 * For k>=1, computing the full quotient in one go would cost
	 * M(kn), which is >= kM(n). Afterwards, recovering the remainder
	 * would cost kM(n) again (mult.kn times n).
	 */
	for(;src->deg>=n;) {
		int m = MIN(n-1,src->deg - n);
		mploc_z2x_reciprocal(a1,src);
		mploc_z2x_chop_deg(a1,m);

		BAD_CODE_WARNING;
		/* This should be a short product. */
		mploc_z2x_mul(t,a1,FIELD_PREINV(a->f));

		mploc_z2x_chop_deg(t,m);
		mploc_z2x_reversal(q,t,m);
		mploc_z2x_mul(t,q,mplf_z2n_def_poly(a->f));
		mploc_z2x_mul_xi(t,t,src->deg - n - m);
		mploc_z2x_sub(src,src,t);
	}

	mploc_z2n_set(r,src);

	mploc_z2x_clear(a1);
	mploc_z2x_clear(q);
	mploc_z2x_clear(t);
	mploc_z2x_clear(src);
}

#if 0
static void mplf_z2n__preinv_reducer_reduce_svoboda(mploc_z2n_ptr r, mploc_z2n_srcptr a)
{
}
#endif

static void mplf_z2n__preinv_reducer_hook(mplf_t f, struct z2x_reducer_t * rf)
{
	ASSERT(f->priv->reducer==NULL);
	f->priv->reducer=rf;
}

static void mplf_z2n__preinv_reducer_unhook(mplf_t f)
{
	struct z2x_reducer_t * rf = f->priv->reducer;
	ASSERT(rf!=NULL);
	ASSERT(rf->magic==MPLF_Z2N__PREINV_REDUCER_MAGIC);
	f->priv->reducer=NULL;

	/* XXX Now, what do we do with *rf ? */
	if (PREINV_REDUCER_PRIV(rf)->autofree) {
		mplf_z2n__preinv_reducer_delete(rf);
	}
}

static void mplf_z2n__preinv_reducer_delete(struct z2x_reducer_t * rf)
{
	mploc_z2x_clear(PREINV_REDUCER_PRIV(rf)->preinv);
#if 1
	mploc_z2x_clear(PREINV_REDUCER_PRIV(rf)->svoboda);
#endif
	free(rf->priv);
	free(rf);
}

static struct z2x_reducer_t * mplf_z2n__preinv_reducer_new(mplf_t f)
{
	struct z2x_reducer_t * rf;
	mplf_t z2=mplf_z2n_def_poly(f)->coeff_field;

	rf = malloc(sizeof(struct z2x_reducer_t));
	rf->magic=MPLF_Z2N__PREINV_REDUCER_MAGIC;
	rf->reduce=mplf_z2n__preinv_reducer_reduce;
	rf->hook=mplf_z2n__preinv_reducer_hook;
	rf->unhook=mplf_z2n__preinv_reducer_unhook;
	rf->handles_overlaps=1;
	rf->priv=malloc(sizeof(struct mplf_z2n__preinv_reducer_priv_t));
	mploc_z2x_init(PREINV_REDUCER_PRIV(rf)->preinv,z2);
	PREINV_REDUCER_PRIV(rf)->autofree=0;

	/* This is the easy part. The hard part is the reducing, where
	 * care must be taken so that computations aren't done at the
	 * highest precision.
	 */
	mploc_z2x_preinv(PREINV_REDUCER_PRIV(rf)->preinv, mplf_z2n_def_poly(f));
#if 1
	{
		mploc_z2x_t q0,b0,f1;
		int n = f->degree;
		int n2 = n - (n/2);
		
		mploc_z2x_init(q0,z2);
		mploc_z2x_init(b0,z2);
		mploc_z2x_init(f1,z2);
		mploc_z2x_init(PREINV_REDUCER_PRIV(rf)->svoboda,z2);
		mploc_z2x_reciprocal(f1,mplf_z2n_def_poly(f));
		mploc_z2x_set(q0,PREINV_REDUCER_PRIV(rf)->preinv);
		mploc_z2x_chop_deg(q0,n2);
		mploc_z2x_mul(b0,q0,f1);
		mploc_z2x_reversal(PREINV_REDUCER_PRIV(rf)->svoboda,b0,n+n2);
		mploc_z2x_clear(f1);
		mploc_z2x_clear(b0);
		mploc_z2x_clear(q0);
	}
#endif

	return rf;
}

void mplf_z2n_set_reducer_preinv(mplf_t f)
{
	struct z2x_reducer_t * rf = mplf_z2n__preinv_reducer_new(f);
	PREINV_REDUCER_PRIV(rf)->autofree=1;
	mplf_z2n_set_reducer(f,rf);
}


void mploc_z2n_reduce(mploc_z2n_ptr dst, mploc_z2n_srcptr src) 
{
	mploc_z2x_t tmp;
	mploc_z2x_ptr tmp_ptr=NULL;

	if (dst==src &&
		!(!src->f->priv->reducer ||
			src->f->priv->reducer->handles_overlaps))
	{
		mploc_z2x_init(tmp,src->coeff_field);
		mploc_z2x_realloc(tmp,1,src->f->degree-1,src->prec);
		tmp->deg=src->f->degree-1;
		tmp_ptr=tmp;
	}

	if (src->f->priv->reducer) {
		(*(src->f->priv->reducer->reduce))(tmp_ptr?tmp_ptr:dst,src);
	} else {
		mploc_z2x_div(NULL,tmp_ptr?tmp_ptr:dst,src,src->f->priv->p);
	}

	if (tmp_ptr) {
		mploc_z2x_set(dst,tmp);
		mploc_z2x_clear(tmp);
	}
	ASSERT(mploc_z2x_is_good_looking(dst));
	ASSERT(mploc_z2n_is_reduced(dst));
}

/* These two are duplicates of the z2x analogues. But we prefer to have
 * them separated.
 */
void mploc_z2n_init(mploc_z2n_ptr x, mplf_t f)
{
	x->prec=0;
	x->deg=-1;
	x->alloc=0;
	x->cw=0;
	x->c=NULL;
	x->f=f;
	x->coeff_field=f->priv->p->coeff_field;
#ifdef DEBUG_MPLOC
	x->f->people++;
#endif
}

void mploc_z2n_clear(mploc_z2n_ptr x)
{               
#ifdef DEBUG_MPLOC
        x->f->people--;
#endif  
        free(x->c);
        memset(x,0,sizeof(mploc_z2n_ptr));
}


void mploc_z2n_mul(mploc_z2n_ptr c, mploc_z2n_srcptr a, mploc_z2n_srcptr b)
{
	/* Caveat here: if a or b have very high precision (more than the
	 * precision of the defining polynomial), some time is wasted.
	 */
	mploc_z2x_mul(c,a,b);
	mploc_z2n_reduce(c,c);
}

void mploc_z2n_sqr(mploc_z2n_ptr c, mploc_z2n_srcptr a)
{
	mploc_z2x_sqr(c,a);
	mploc_z2n_reduce(c,c);
}

void mploc_z2n_rshift(mploc_z2n_ptr z, mploc_z2n_srcptr x, mp_size_t n)
{
	if (EXPECT_FALSE(n==0)) {
		mploc_z2n_set(z,x);
	}

	if (!mploc_z2n_is_reduced(x)) {
		mploc_z2n_reduce(z,x);
	}
	/* This *has* to be so ; otherwise result is not well-defined... */
	ASSERT(mploc_z2x_valuation(x)>=n);
	mploc_z2x_rshift(z,x,n);
}

/* Note : We could do the reduction only at the very end instead...
 */
void mploc_z2n_inv_improve_recursive(mploc_z2n_ptr x, mploc_z2n_srcptr a,
		mp_size_t aim)
{
	mploc_z2x_t y;
	mploc_z2x_t t;

	if (x->prec >= aim) return;

	mploc_z2n_init(t,a->f);
	mploc_z2n_init(y,a->f);

	mploc_z2n_inv_improve_recursive(x,a,aim - aim/2);
	mploc_z2n_set_change_precision(t,a,aim);
	mploc_z2n_force_precision(x,aim);
	mploc_z2n_mul(y,t,x);
	BAD_CODE_WARNING;

	/* XXX That's a bad way of making a full product */

	mploc_z2n_sub_ui(y,y,1);
	mploc_z2n_mul(y,y,x);
	mploc_z2n_neg(y,y);
	mploc_z2n_set_change_precision(t,y,aim);
	mploc_z2n_add(x,x,t);

	x->prec=aim;

	mploc_z2n_clear(t);
	mploc_z2n_clear(y);

	ASSERT(mploc_z2x_is_good_looking(x));
}

static const char* call_reduce_insult=
	"Call _reduce() before complicated functions";

void mploc_z2n_inv(mploc_z2n_ptr z, mploc_z2n_srcptr a)
{
	rf_elt r1;
	mploc_z2x_t x;

	ASSERT(mploc_z2n_is_good_looking(a));

	if (!mploc_z2n_is_reduced(a)) {
		BASH_USER(call_reduce_insult);
		/* Otherwise we have to go through unpleaseant bloatware
		 * stuff. And to no benefit, since a is const.
		 */
	}

	if (EXPECT_FALSE(mploc_z2n_valuation(a))) DIVIDE_BY_ZERO();

	r1=new_rfe(a->f);
	rfe_set_mploc(r1,a,a->f);
	RF_inv(r1,r1,a->f);

	mploc_z2n_init(x,a->f);
	mploc_set_rfe(x,r1,a->f);
	delete_rfe(r1,a->f);
	mploc_z2n_inv_improve_recursive(x,a,a->prec);

	mploc_z2n_set(z,x);
	mploc_z2n_clear(x);
}

static void mploc_z2n_invsqrt_improve_recursive(
		mploc_z2n_ptr x,
		mploc_z2n_srcptr a,
		mp_size_t aim)
{
	mploc_z2x_t y,t;

	/* Let x be the value to reach, and xtilde the current
	 * approximation. Let e = xtilde-x be the current error.
	 * We assume that e=O(2^k) (k is x->prec, here).
	 *
	 * We have:
	 * (xtilde/x)^2 = 1 + 2e/x + O(2^(2k))
	 * xtilde.((xtilde/x)^2-1)/2
	 *            = e/x . xtilde + O(2^(2k-1))
	 *            = e + O(2^(2k-1))
	 *
	 * This means that from input precision k, we obtain 2k-1 (hence
	 * k must be >= 2, cf the artin-schreier root computation below).
	 *
	 * to obtain (aim), we must start with k sufficiently big. This
	 * means:
	 * 	2k-1 >= aim
	 * 	2k >= aim+1
	 * 	k >= ceil((aim+1)/2) = (aim+1) - floor((aim+1)/2)
	 *
	 * (usual deal n = ceil(n/2) + floor(n/2) for any integer n).
	 */

	if (x->prec >= aim) return;
	mploc_z2n_init(t,a->f);
	mploc_z2n_init(y,a->f);
	mploc_z2n_invsqrt_improve_recursive(x,a,aim+1 - (aim+1)/2);

	/* One must *not* change the precision of a here ... */
	mploc_z2n_set(t,a);
	ASSERT(a->prec >= aim + 1);

	mploc_z2n_force_precision(x,aim);
	mploc_z2x_sqr(y,x);
	mploc_z2x_mul(y,y,t);
	mploc_z2n_sub_ui(y,y,1);
	/* Note that here, mploc_z2n_rshift performs the reduction
	 * itself.
	 */
	mploc_z2n_rshift(y,y,1);
	mploc_z2n_mul(y,y,x);
	mploc_z2n_set_change_precision(t,y,aim);
	mploc_z2n_sub(x,x,t);
	x->prec=aim;
	mploc_z2n_clear(t);
	mploc_z2n_clear(y);

	ASSERT(mploc_z2x_is_good_looking(x));
}

int mploc_z2n_invsqrt(mploc_z2n_ptr z, mploc_z2n_srcptr a)
{
	mploc_z2x_t x,y;
	rf_elt xbar,h;
	int r;

	ASSERT(mploc_z2n_is_good_looking(a));

	if (!mploc_z2n_is_reduced(a)) {
		BASH_USER(call_reduce_insult);
	}

	if (EXPECT_FALSE(mploc_z2n_valuation(a))) DIVIDE_BY_ZERO();

	xbar=new_rfe(a->f);

	rfe_set_mploc(xbar,a,a->f);
	RF_invsqrt(xbar,xbar,a->f);

	if (a->prec==1) {
		mploc_set_rfe(z,xbar,a->f);
		delete_rfe(xbar,a->f);
		/* Two returns codes are possible, based on different
		 * interpretations.
		 *
		 * return 0 : we cannot decide on the existence of the
		 * square root based on the single knowledge of the first
		 * bit. Indeed, 1+2, for instance, can have no (inverse)
		 * square root in an unramified extension of Z2, while
		 * its square root in Z2[r], where r^2-2=0, is 1+r. And
		 * Z2[r] is (wildly) ramified.
		 *
		 * return 1 : we think about it as something happening in
		 * the residue field only. Therefore the sqrt operation
		 * should be successful.
		 */
		return 1;
	}

	mploc_z2n_init(x,a->f);
	mploc_set_rfe(x,xbar,a->f);

	/* Is the square root lying in a ramified extension ? */
	mploc_z2n_force_precision(x,2);
	mploc_z2n_init(y,a->f);
	mploc_z2n_sqr(y,x);
	mploc_z2n_mul(y,y,a);
	mploc_z2n_sub_ui(y,y,1);
	r=mploc_z2n_valuation(y);

	ASSERT(r>=1);

	if (r<2) {
		/* The result has a short valuation. This means that a
		 * ramified extension must be taken to have the result.
		 */
		mploc_z2n_clear(y);
		mploc_z2n_clear(x);
		delete_rfe(xbar,a->f);
		return 0;
	}
	
	if (a->prec==2) {
		/* Then this is all right. We have our square root in an
		 * unramified extension
		 */
		mploc_z2n_set(z,x);
		mploc_z2n_clear(y);
		mploc_z2n_clear(x);
		delete_rfe(xbar,a->f);
		return 1;
	}
	
	/* Let x be the targeted solution, and x0=xbar the approximation
	 * at order 1. We have x^2a=1, x0^2a=1 mod 2.
	 * 
	 * Let x0=x+2e+O(4)
	 *
	 * We have (x0/x)^2 = (1 + (2e/x) + O(4))^2
	 *                  = 1 + 4((e/x)+(e/x)^2) + O(8)
	 *                  = 1 + 4h + O(8)
	 * Therefore h should be such that X^2+X-h has a root in F_{2^n}
	 */

	/* We are actually computing -h */
	h=new_rfe(a->f);
	rfe_set_mploc_coeff(h,y,2,a->f);
	r=RF_has_trace_zero(h,a->f);

	mploc_z2n_clear(y);

	if (!r) {
		mploc_z2n_clear(x);
		delete_rfe(h,a->f);
		delete_rfe(xbar,a->f);
		return 0;
	}
	
	RF_artin_schreier_root(h,h,a->f);
	RF_mul(h,h,xbar,a->f);
	/* Here, we have recovered the error coefficient e from the
	 * formulas above.
	 */

	/* This coefficient belongs to the current approximation, set it.
	 * NOTE: this automatically updates x->prec
	 */
	mploc_coeff_set_rfe(x,1,h,a->f);
	delete_rfe(xbar,a->f);
	delete_rfe(h,a->f);

	/* We're ready to roll the newton iteration */
	mploc_z2n_invsqrt_improve_recursive(x,a,a->prec-1);
	mploc_z2n_set(z,x);
	mploc_z2n_clear(x);

	return r;
}

int mploc_z2n_sqrt(mploc_z2n_ptr z, mploc_z2n_srcptr a)
{
	mp_size_t v;
	int r;

	if (a->deg==-1) {
		mploc_z2n_set_change_precision(z,a,MAX(a->prec-(a->prec>1),0));
		return 1;
	}

	v=mploc_z2n_valuation(a);

	if (v & 1)
		return 0;

	if (a==z || v) {
		mploc_z2n_t tmp;

		mploc_z2n_init(tmp,a->f);
		mploc_z2n_rshift(tmp,a,v);
		r=mploc_z2n_invsqrt(z,tmp);
		if (r) {
			mploc_z2n_mul(z,z,tmp);
			mploc_z2n_lshift(z,z,v/2);
		}
		mploc_z2n_clear(tmp);
	} else {
		r=mploc_z2n_invsqrt(z,a);
		if (r) {
			mploc_z2n_mul(z,z,a);
		}
	}
	return r;
}

int mploc_z2n_eq(mploc_z2n_srcptr x, mploc_z2n_srcptr y)
{
	int res;

	if (!mploc_z2n_is_reduced(y) || !mploc_z2n_is_reduced(x)) {
		mploc_z2x_t t;
		mploc_z2n_init(t,x->f);
		mploc_z2n_sub(t,x,y);
		mploc_z2n_reduce(t,t);
		res=mploc_z2x_eq_ui(t,0);
		mploc_z2n_clear(t);
	} else {
		res=mploc_z2x_eq(x,y);
	}

	return res;
}

int mploc_z2n_eq_ui(mploc_z2n_srcptr x, mp_limb_t u)
{
	int res;

	if (!mploc_z2n_is_reduced(x)) {
		mploc_z2x_t t;
		mploc_z2n_init(t,x->f);
		mploc_z2n_sub_ui(t,t,u);
		mploc_z2n_reduce(t,t);
		res=mploc_z2x_eq_ui(t,0);
		mploc_z2n_clear(t);
	} else {
		res=mploc_z2x_eq_ui(x,u);
	}

	return res;
}

/* Comp ute the trace as the trace of the multiplication-by-x map in base
 * 1,t,t^2,...,t^(n-1)
 *
 * The reduction function gets called at each step. Hopefully it is fast
 * when the input degree is n...
 *
 */
void mploc_z2n_trace(mploc_ptr z, mploc_z2n_srcptr x)
{
	mploc_t r;
	mploc_z2x_t p;
	int i;

	mploc_z2_init(r,z->f);
	mploc_z2n_init(p,x->f);

	mploc_z2_set_ui(z,0,x->prec);
	mploc_z2n_set(p,x);
	for(i=0;i<x->f->degree;i++) {
		mploc_z2x_get_coeff(r,p,i);
		mploc_z2_add(z,z,r);
		mploc_z2x_mul_xi(p,p,1);
		mploc_z2n_reduce(p,p);
	}

	mploc_z2n_clear(p);
	mploc_z2_clear(r);
}

