/* $Id: z2x.c 203 2005-01-19 15:02:49Z /CN=gaudry/emailAddress=gaudry@polytechnique.fr $ */

/***************************************************************************/
/* HEADERS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* This is all the GMP stuff I need... */
#include <gmp.h>


#ifdef	HAVE_CONFIG_H 
#include <config.h>
#endif

#include <assert.h>

#include "gmp_hacks.h"
#include "mploc.h"
#include "z2_limbs.h"
#include "z2.h"
#include "asm/asm.h"
#include "z2x.h"

#include "manu.h"

#define BAD_CODE_WARNING				\
	BEGIN_BLOCK					\
		static int i;				\
		if (!i++) {				\
			WARNING("This is bad code");	\
		}					\
	END_BLOCK


/* ATTENTION : No underlying finite field implementation is called. */

struct mplf_private {
	/* NOTHING FOR NOW... */
};

/***************************************************************************/
/* ENTRY POINTS -- POLYNOMIAL INTERFACE */

/* Initialize a polynomial (initialized with degree -1, precision 0, no
 * allocated coeffs.
 *
 * Note that f is mostly useless, but we use it for population count on
 * init and close (one poly counts as 1)
 */
void mploc_z2x_init(mploc_z2x_ptr x, mplf_t f UNUSED_VARIABLE)
{
	x->prec=0;
	x->deg=-1;
	x->alloc=0;
	x->c = NULL;
	x->cw = 0;
#ifdef DEBUG_MPLOC
	x->f=NULL;
	x->coeff_field=f;
	x->coeff_field->people++;
#endif
}

/* frees up the poly */
void mploc_z2x_clear(mploc_z2x_ptr x)
{
#ifdef DEBUG_MPLOC
	x->coeff_field->people--;
#endif
	free(x->c);
	memset(x,0,sizeof(mploc_z2x_ptr));
}

int mploc_z2x_is_good_looking_maybe_nonnormal(mploc_z2x_srcptr x)
{
	if (x->alloc < (x->deg + 1)*x->cw) {
		fprintf(stderr, "Poly has unallocated coeffs\n");
		return 0;
	}
	if (x->cw < iceildiv(x->prec,GMP_LIMB_BITS)) {
		fprintf(stderr, "Poly has too large precision\n");
		return 0;
	}
	if (x->deg<-1) {
		fprintf(stderr, "Poly has unexpected degree (<-1)\n");
		return 0;
	}
#ifdef	DEBUG_MPLOC
	if (x->coeff_field == NULL) {
		fprintf(stderr, "Poly has no coeff field\n");
		return 0;
	}
	if (x->coeff_field->people <= 0) {
		fprintf(stderr, "Poly has coeff field with <=0 elts.\n");
		return 0;
	}
#endif

#ifndef	LOOSE_CW
	if (x->cw != iceildiv(x->prec,GMP_LIMB_BITS) && !(x->prec==0 && x->cw==1)) {
		fprintf(stderr, "Poly has bad cw\n");
		return 0;
	}
#endif
	return 1;
}

int mploc_z2x_is_good_looking(mploc_z2x_srcptr x)
{
	if (!mploc_z2x_is_good_looking_maybe_nonnormal(x)) {
		return 0;
	}
	if (x->prec==0 && x->deg!=-1) {
		fprintf(stderr, "Poly has precision 0 and degree>=0\n");
		return 0;
	}
	if (x->deg>-1 && mploc_limbs_valuation(x->c+x->deg*x->cw,x->cw)>=x->prec) {
		fprintf(stderr, "Poly has zero leading coeff\n");
		return 0;
	}
	return 1;
}

/*
 *
 * Recall that a mploc_z2x_t contains the following fields:
 * 	mp_size_t       prec;
 * 	int             deg;
 * 	int             alloc;
 * 	mp_size_t       cw;     
 * 				
 * It means the following:
 *   prec   is the precision (in bits) at which the coefficients of the
 *          polynomials make sense.
 *   deg    is the degree of the polynomial (i.e. there are deg+1 coeffs)
 *   alloc  is the memory space that has been allocated to store the
 *          coefficients. The unit, here, is one limb.
 *   cw     is the number of words used to store one coefficient.
 * 
 * According to whether LOOSE_CW is defined, cw is tight or not, when an
 * element is in the "standard" state (meaning when entering or leaving
 * one of the high level functions).
 * 
 * Inside internal functions of z2x.c, a mploc_z2x_t can be in various
 * states:
 *   Standard
 *   UnNormalized
 * 
 * In the Standard state, the following predicates that should be true:
 *   (deg == -1)   ||   (deg >= 0)
 *   cw*GMP_LIMB_BITS >= prec
 *   alloc >= cw*(deg+1)
 *   alloc == Size of the allocated bloc c  
 *   LeadingCoeff is not zero (if deg >= 0)
 *   prec == number of meaningful bits in each coeff
 * Furthermore, if LOOSE_CW is not defined, cw is the smallest possible.
 * 
 * In the UnNormalized state, we have:
 *   (deg == -1)   ||   (deg >= 0)
 *   cw*GMP_LIMB_BITS >= prec
 *   alloc >= cw*(deg+1)
 *   alloc == Size of the allocated bloc c
 *   prec == number of meaningful bits in each coeff
 * NB: the leadingcoefficient might be zero. On the other hand, quite
 * often in that state there is meaningful data that is stored in places
 * higher than the degree.
 * HENCE: UnNormalized essentially means that the degree is not yet set
 * (with the restriction that it is only allowed to stay in accordance
 * with the allocated space for coeffs)
 * 
 * *** mploc_z2x_normalize
 * Takes an mploc_z2x_t in UnNormalized state and put it in Standard
 * state by reducing the degree until the leadingcoeff is 0 or the
 * polynomial is 0.  Only deg is changed (and decreased).
 * WARNING: it might be necessary to put an upper bound on the degree
 * before calling this function.
 * 
 * *** mploc_z2x_clobber
 * Set the element to 0 by just putting deg to -1 and prec to 0.  If we
 * are not in LOOSE_CW, the output is neither Standard nor UnNormalized,
 * since cw is not adjusted.
 * 
 * *** mploc_z2x_realloc  (case dir == 1)
 * On input, it takes an element that is Standard, but that can have a
 * non-adjusted cw. It also takes deg and prec.
 * 
 * The output is an element that is in UnNormalized state.  The function
 * works by side-effects, but for simplicity, we call x the input and y
 * the output.
 * 
 * Getting out of the function, y is such that:
 *   y->prec = prec
 *   y->cw is enough for that prec
 *   y->deg is unchanged, BUT: if we put y->deg=deg therafter, the
 *     element will be UnNormalized state (meaning that there is enough
 *     room for storing that much elements, each on y->cw limbs). 
 *   y->alloc is set accordingly
 *   The coefficients of x up to x->deg are copied into y, but the precision
 *   of the coeffs is forced to prec.

 * In short:
 *
 * A shrink/extension must be thoughtful (sorry), and done in typically three
 * steps.
 *
 * 1) Clobber the input if possible (mploc_z2x_clobber is the trivial
 *    function below). Whether this is doable or not can only be decided
 *    by the caller.
 * 2) Call mploc_z2x_realloc with the appropriate args. maybe you will
 *    have to change ``deg'' to MAX(deg,x->deg) if you want data to be
 *    kept.
 * 3) Change x->deg ; This step may have to be done later.
 *
 * A number of ASSERTs here and there make sure that things are done this
 * way.
 */

void mploc_z2x_clobber(mploc_z2x_ptr x)
{
	x->deg=-1;
	x->prec=0;
}

void mploc_z2x_aggressive_adjust_cw(mploc_z2x_ptr x)
{
	ASSERT(x->deg==-1);
	x->cw=iceildiv(x->prec,GMP_LIMB_BITS);
}

int mploc_z2x_realloc(mploc_z2x_ptr x,
		int dir,
		mp_size_t deg,
		mp_size_t prec)
{
	mp_size_t nw, needed_space;
	int i;
	int need_realloc, need_change_cw;

	ASSERT(dir==-1 || dir==1);
	ASSERT(deg>=-1);
	ASSERT(prec>=0);

	ASSERT(deg>=x->deg);
	/* We do not allow chopping off coefficients aggressively. The
	 * caller must call z2x_clobber() before.
	 * This way, we can keep track of the data present in x.
	 */
	
	nw = iceildiv(prec,GMP_LIMB_BITS);

	needed_space = (deg+1)*nw;
	need_realloc = (dir>0 && x->alloc < needed_space);

	ASSERT(!(dir<0 && nw > x->cw));
	/* This means that the dir flag is not properly set. */

#ifdef	LOOSE_CW
	need_change_cw = (dir > 0 && nw > x->cw) || (dir < 0 && nw < x->cw);

	/* We may have to _shorten_ cw even if dir>0 ; there may be
	 * enough space to accomodate (deg+1)*nw, but not (deg+1)*x->cw
	 *
	 * Such a condition does not exist for deg < 0, of course (if nw
	 * and x->cw are different, nw will be the new value on exit)
	 */
	if (dir>0 && !need_realloc && (deg+1)*x->cw > x->alloc) {
		/* Then no realloc, for sure, but we _do_ have to change
		 * cw ; and from the conditions above, need_change_cw
		 * can't be set yet.
		 */
		need_change_cw=1;
	}
#else
	need_change_cw = nw != x->cw;
#endif

	/* At this point, need_change_cw and need_realloc are correct */

	if (!need_realloc && !need_change_cw) {
		ASSERT(x->alloc >= (deg+1)*x->cw);
		x->prec=prec;
		return 0;
	}

	if (need_realloc) {
		mp_limb_t * nc;
		mp_size_t na;
		mp_size_t mw;

		MEMSIZE_NEXT(na,needed_space,x->alloc);
		LOG_REALLOC();
		nc = malloc(na*sizeof(mp_limb_t));

		/* Since we don't do realloc() anymore, we have to handle
		 * the data copy stuff. It's perhaps a pity not to do a
		 * simple memcpy in the (common) case where nw == x->cw ;
		 * FIXME (investigate, bench, fix). */
		mw=MIN(nw,x->cw);
		for(i=0;i<=x->deg;i++)
			mploc_limbs_copy(nc+i*nw,x->c+i*x->cw,mw);
		free(x->c);
		x->c=nc;
		x->alloc=na;
	} else if (x->cw && need_change_cw) {
		LOG_SHUFFLE();
		if (nw < x->cw) {
			for(i=0;i<=x->deg;i++)
				mploc_limbs_move(x->c+i*nw,x->c+i*x->cw,nw);
		} else if (nw > x->cw) {
			for(i=x->deg;i>0;i--)
				mploc_limbs_move(x->c+i*nw,x->c+i*x->cw,x->cw);
		}
	}

	x->cw=nw;
	x->prec=prec;

	ASSERT(x->alloc >= (deg+1)*x->cw);

	/* Note that x->deg is not changed, knowingly */

	ASSERT(mploc_z2x_is_good_looking_maybe_nonnormal(x));

	return 1;
}

/* This is a pretty trivial application of the _realloc() interface() */

/* Change the precision, extending if necessary. See discussion above.
 * x->prec *is changed* by this operation, but high bits are garbage
 */
void mploc_z2x_force_precision(mploc_z2x_ptr x, mp_size_t k)
{
	if (k<x->prec) {
		mploc_z2x_realloc(x,1,x->deg,k);
		mploc_z2x_normalize(x);
	} else {
#ifdef	VALGRIND_FIXUPS
		mp_size_t oprec=x->prec;
#endif
		mploc_z2x_realloc(x,1,x->deg,k);
#ifdef	VALGRIND_FIXUPS
		mploc_z2x_sign_extend(x,oprec);
#endif
	}
}

/* Chop off all coefficients above given degree */
void mploc_z2x_chop_deg(mploc_z2x_ptr x, mp_size_t deg)
{
	ASSERT(deg>=-1);
	x->deg=MIN(x->deg,deg);
	mploc_z2x_normalize(x);
}

/* Remove the leading zero coefficients in x */
void mploc_z2x_normalize(mploc_z2x_ptr z)
{
	mp_size_t aw = TIGHTCW(z);

#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	for(;z->deg>=0;z->deg--) {
		if (mploc_limbs_valuation(z->c+z->deg*z->cw,aw)<z->prec)
			break;
	}

	ASSERT(mploc_z2x_is_good_looking(z));
}

/* Chop off all the bits above k -- This can be done after a precision
 * extension, starting from precision k, in order to ``pad with zeroes''
 */
void mploc_z2x_insert_zero(mploc_z2x_ptr x, mp_size_t k)
{
	int i;
	mp_size_t aw = TIGHTCW(x);

#ifndef	LOOSE_CW
	ASSERT(aw==x->cw);
#endif

	for(i=0;i<=x->deg;i++) {
		mploc_limbs_insert_zero(x->c+i*x->cw,aw,k);
	}

	/* 20040825 ; normalize() inserted on a suggestion of TH.
	 *
	 * TO DO : Check that _insert_zero never gets used in any
	 * critical path.
	 *
	 * If _insert_zero is to be public to the user, normalization is
	 * mandatory.
	 *
	 * If not, then it shouldn't exist since it isn't used.
	 *
	 * If it becomes used, and tagged private, I probably could get
	 * along with is_good_looking_maybe_nonnormal and no
	 * normalization.
	 */
	if (k<x->prec) mploc_z2x_normalize(x);

	ASSERT(mploc_z2x_is_good_looking(x));
}

void mploc_z2x_sign_extend(mploc_z2x_ptr x, mp_size_t k)
{
	int i;
	mp_size_t aw = TIGHTCW(x);
#ifndef LOOSE_CW
	ASSERT(aw==x->cw);
#endif
	for(i=0;i<=x->deg;i++) {
		mploc_limbs_sign_extend(x->c+i*x->cw,aw,k);
	}
	if (k<x->prec) mploc_z2x_normalize(x);
	ASSERT(mploc_z2x_is_good_looking(x));
}

/* x <- y + O(2^k) */
void mploc_z2x_set_change_precision(mploc_z2x_ptr x, mploc_z2x_srcptr y, mp_size_t k)
{
	mp_size_t aw;
	
	if (EXPECT_FALSE(x==y)) {
		mploc_z2x_force_precision(x,k);
		/* result is normalized above */
		return;
	}

	aw = iceildiv(k,GMP_LIMB_BITS);

	ASSERT(x!=y);

	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,y->deg,k);

#ifndef	LOOSE_CW
	mploc_z2x_aggressive_adjust_cw(x);
#endif

	x->deg=y->deg;

#ifdef	VALGRIND_FIXUPS
	mploc_limbs_zero(x->c,(y->deg+1)*x->cw);
#endif

	if (x->cw != y->cw) {
		mp_size_t i;
		for(i=0;i<=y->deg;i++) {
			mploc_limbs_copy(x->c+i*x->cw,y->c+i*y->cw,MIN(aw,y->cw));
		}
	} else {
		mploc_limbs_copy(x->c,y->c,(y->deg+1)*y->cw);
	}

	mploc_z2x_normalize(x);

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* x <- y */
void mploc_z2x_set(mploc_z2x_ptr x, mploc_z2x_srcptr y)
{
	mploc_z2x_set_change_precision(x,y,y->prec);
}

/* x <- X^i + O(2^prec) */
void mploc_z2x_set_xi(mploc_z2x_ptr x, mp_size_t i, mp_size_t prec)
{
	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,i,prec);
	x->deg=i;

	mploc_limbs_zero(x->c,(i+1)*x->cw);
	x->c[x->cw*i]=1UL;

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* [X^i] x <- u + O(2^prec)
 *
 * If gprec is the precision of other elements, gprec <- MIN(prec,gprec)
 */
void mploc_z2x_set_coeff(mploc_z2x_ptr x, mp_size_t i, mploc_srcptr a)
{
	mploc_z2x_realloc(x,1,MAX(x->deg,i),MIN(x->prec,a->prec));
	if (i>x->deg+1)
		mploc_limbs_zero(x->c+(x->deg+1)*x->cw,(i-1-x->deg)*x->cw);
	x->deg=MAX(x->deg,i);

	mploc_limbs_copy(x->c+i*x->cw,PTR(a->x),MIN(x->cw,SIZ(a->x)));

	/* A whole bunch of funny things can happen... Better to
	 * normalize here. */
	mploc_z2x_normalize(x);

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* [X^i] x <- u + O(2^prec)
 *
 * If gprec is the precision of other elements, gprec <- MIN(prec,gprec)
 */
void mploc_z2x_set_coeff_ui(mploc_z2x_ptr x, mp_size_t i, mp_limb_t u)
{
	mploc_z2x_realloc(x,1,MAX(x->deg,i),x->prec);
	if (i>x->deg+1)
		mploc_limbs_zero(x->c+(x->deg+1)*x->cw,(i-1-x->deg)*x->cw);
	x->deg=MAX(x->deg,i);

	mploc_limbs_zero(x->c+i*x->cw+1,x->cw-1);
	x->c[i*x->cw]=u;

	/* A whole bunch of funny things can happen... Better to
	 * normalize here. */
	mploc_z2x_normalize(x);

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* u <- [X^i] x */
void mploc_z2x_get_coeff(mploc_ptr a, mploc_z2x_srcptr x, mp_size_t i)
{
	ASSERT(i>=0);

	mploc_z2_force_precision(a,x->prec);
	if (i<=x->deg)
		mploc_limbs_copy(PTR(a->x),x->c+i*x->cw,SIZ(a->x));
	else
		mploc_limbs_zero(PTR(a->x),SIZ(a->x));

	ASSERT(mploc_z2_is_good_looking(a));
}

/* x <- u + O(2^prec) */
void mploc_z2x_set_ui(mploc_z2x_ptr x, mp_limb_t u, mp_size_t prec)
{
	mp_size_t aw;
	
	if (EXPECT_FALSE(!u || prec==0 ||
		(prec < GMP_LIMB_BITS && !(u & ((1UL<<prec)-1))))) {
		mploc_z2x_clobber(x);
		mploc_z2x_realloc(x,1,-1,prec);
		x->deg=-1;
		ASSERT(mploc_z2x_is_good_looking(x));
		return;
	}

	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,0,prec);
	x->deg=0;

	aw = TIGHTCW(x);

#ifndef	LOOSE_CW
	ASSERT(aw==x->cw);
#endif

	mploc_limbs_zero(x->c,aw);
	x->c[0]=u;

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* x <- a + O(2^prec) */
void mploc_z2x_set_mpz(mploc_z2x_ptr x, mpz_srcptr a, mp_size_t prec)
{
	mp_size_t aw;
	mp_size_t s;

	if (!SIZ(a)) {
		mploc_z2x_clobber(x);
		mploc_z2x_realloc(x,1,-1,prec);
		x->deg=-1;
		return;
	}

	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,0,prec);
	x->deg=0;
	
	aw = TIGHTCW(x);

#ifndef	LOOSE_CW
	ASSERT(aw==x->cw);
#endif

	s=ABS(SIZ(a));

	if (s>aw) {
		mploc_limbs_copy(x->c,PTR(a),aw);
	} else {
		mploc_limbs_copy(x->c,PTR(a),SIZ(a));
		mploc_limbs_insert_zero(x->c,aw,prec);
	}

	if (SIZ(a)<0) {
		mploc_limbs_neg(x->c,x->c,aw);
	}

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* x <- Random poly of degree d and precision k
 *
 * There's a bit of code duplication here. (GMP's fault)
 */
void mploc_z2x_set_random(mploc_z2x_ptr x, gmp_randstate_t prng,
		mp_size_t d, mp_size_t prec)
{
	mpz_t tmp;

	mploc_z2x_clobber(x);
	mploc_z2x_realloc(x,1,d,prec);
	mploc_z2x_aggressive_adjust_cw(x);
	x->deg=d;
	mpz_init2(tmp,(x->deg+1)*x->cw*GMP_LIMB_BITS);
	mpz_urandomb(tmp,prng,(x->deg+1)*x->cw*GMP_LIMB_BITS);
	mploc_limbs_copy(x->c,PTR(tmp),(x->deg+1)*x->cw);
	mpz_clear(tmp);

	mploc_z2x_normalize(x);

	ASSERT(mploc_z2x_is_good_looking(x));
}

/* z <- x + y */
void mploc_z2x_add(mploc_z2x_ptr z, mploc_z2x_srcptr x, mploc_z2x_srcptr y)
{
	mp_size_t aw;
	mp_size_t ad;
	mp_size_t i;

	ad = MAX(x->deg,y->deg);
	if (z!=x && z!=y) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,ad,MIN(x->prec,y->prec));
	/* DO NOT change z->deg to ad now (in case z==x or z==y) */
	
	aw = TIGHTCW(z);

#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	for(i=0;i<=MIN(x->deg,y->deg);i++) {
		mpn_add_n(z->c+i*z->cw,x->c+i*x->cw,y->c+i*y->cw,aw);
	}
	if (x->deg > y->deg) {
		for(;i<=ad;i++) {
			mploc_limbs_move(z->c+i*z->cw,x->c+i*x->cw,aw);
		}
	} else if (x->deg < y->deg) {
		for(;i<=ad;i++) {
			mploc_limbs_move(z->c+i*z->cw,y->c+i*y->cw,aw);
		}
	}

	/* Do the degree change here in case z=x and deg(y)>deg(x) !  */
	z->deg = ad;

	mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}

/* Add the integer u to constant coefficient */
void mploc_z2x_add_ui(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_limb_t u)
{
	mp_size_t aw;
	mploc_z2x_set(z,x);
	if (EXPECT_FALSE(u==0)) return;

	aw = TIGHTCW(z);
	
#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	mploc_z2x_realloc(z,1,MAX(z->deg,0),z->prec);
	mpn_add_1(z->c,z->c,aw,u);
	z->deg=MAX(z->deg,0);

	if (z->deg==0) mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}

/* z <- x - y */
void mploc_z2x_sub(mploc_z2x_ptr z, mploc_z2x_srcptr x, mploc_z2x_srcptr y)
{
	mp_size_t aw;
	mp_size_t ad;
	mp_size_t i;

	ad = MAX(x->deg,y->deg);
	if (z!=x && z!=y) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,ad,MIN(x->prec,y->prec));
	/* DO NOT change z->deg to ad now (in case z==x or z==y) */
	
	aw = TIGHTCW(z);

#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	for(i=0;i<=MIN(x->deg,y->deg);i++) {
		mpn_sub_n(z->c+i*z->cw,x->c+i*x->cw,y->c+i*y->cw,aw);
	}
	if (x->deg > y->deg) {
		for(;i<=ad;i++) {
			mploc_limbs_move(z->c+i*z->cw,x->c+i*x->cw,aw);
		}
	} else if (x->deg < y->deg) {
		for(;i<=ad;i++) {
			mploc_limbs_neg(z->c+i*z->cw,y->c+i*y->cw,aw);
		}
	}

	/* Do the degree change here in case z=x and deg(y)>deg(x) !  */
	z->deg = ad;

	mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}

/* Subtract integer u to the constant coeff */
void mploc_z2x_sub_ui(mploc_z2x_ptr z, mploc_z2x_srcptr x, unsigned long u)
{
	mp_size_t aw;
	mploc_z2x_set(z,x);
	aw = TIGHTCW(z);
	
#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	mpn_sub_1(z->c,z->c,aw,u);
	
	if (x->deg==0) mploc_z2x_normalize(z);
	
	ASSERT(mploc_z2x_is_good_looking(z));
}

void mploc_z2x_neg(mploc_z2x_ptr z, mploc_z2x_srcptr x)
{
	mp_size_t i;
	mp_size_t axw;

	ASSERT(mploc_z2x_is_good_looking(z));
	ASSERT(mploc_z2x_is_good_looking(x));

	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,x->deg,x->prec);
	z->deg=x->deg;

	axw = TIGHTCW(x);

	ASSERT(z->cw>=axw);

	for(i=0;i<=x->deg;i++) {
		mploc_limbs_neg(z->c+i*z->cw,x->c+i*x->cw,axw);
	}
	ASSERT(mploc_z2x_is_good_looking(z));
}


/* z <- X^i * x */
void mploc_z2x_mul_xi(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t i)
{
	mp_size_t aw;
	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,x->deg+i,x->prec);
	/* z->deg changed later */

	/* 20040824 ; Wild fix ; check */
	if (x->deg==-1) {
		z->deg=-1;
		return;
	}

	aw = TIGHTCW(z);

#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	if (z->cw != x->cw) {
		mp_size_t j;
		for(j=MAX(-i,0);j<=x->deg;j++) {
			mploc_limbs_copy(z->c+(i+j)*z->cw,x->c+j*x->cw,aw);
		}
	} else {
		/* We must use move if z==x */
		if (i>=0) {
			mploc_limbs_move(z->c+i*z->cw,x->c,(x->deg+1)*x->cw);
		} else {
			mploc_limbs_move(z->c,x->c+(-i)*x->cw,(x->deg+i+1)*x->cw);
		}
	}

	mploc_limbs_zero(z->c,MAX(i,0)*x->cw);

	z->deg = x->deg + i;

	ASSERT(mploc_z2x_is_good_looking(z));
}

void mploc_z2x_reversal(mploc_z2x_ptr dst, mploc_z2x_srcptr src, mp_size_t deg)
{
	int i;
	mp_size_t scw;

	ASSERT(deg>=src->deg);

	if (EXPECT_FALSE(dst==src)) {
		mploc_z2x_t tmp;
		mploc_z2x_init(tmp, src->coeff_field);
		mploc_z2x_reversal(tmp,src,deg);
		mploc_z2x_set(dst,tmp);
		mploc_z2x_clear(tmp);
		return;
	}
	mploc_z2x_clobber(dst);
	mploc_z2x_realloc(dst,1,deg,src->prec);
	scw=TIGHTCW(src);
	for(i=0;i<=src->deg;i++) {
		mploc_limbs_copy(dst->c+(deg-i)*dst->cw, src->c+i*src->cw, scw);
	}
	mploc_limbs_zero(dst->c,(deg-src->deg)*dst->cw);
	dst->deg=deg;
	mploc_z2x_normalize(dst);
}

/* z <- x div X^i */
void mploc_z2x_div_q_xi(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t i)
{
	if (i>x->deg) {
		/* The precision is x->prec because the whole polynomial
		 * is understood modulo 2^prec */
		mploc_z2x_set_ui(z,0,x->prec);
	} else {
		mploc_z2x_mul_xi(z,x,-i);
	}
}

/* z -> x mod X^i */
void mploc_z2x_div_r_xi(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t i)
{
	mploc_z2x_set(z,x);
	if (i>x->deg)
		return;

	z->deg=i-1;
	
	mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}

/* z -> a * x */
void mploc_z2x_mul_z2(mploc_z2x_ptr z, mploc_z2x_srcptr x, mploc_srcptr a)
{
	int i=0;
	mp_limb_t * buf;

	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,x->deg,MIN(x->prec,a->prec));
	z->deg=x->deg;

	buf = FAST_ALLOC(2 * z->cw * sizeof(mp_limb_t));

	for(i=0;i<=x->deg;i++) {
		mploc_limbs_short_product(z->c+i*z->cw,
				x->c+i*x->cw,PTR(a->x),buf,z->cw);
	}
	
	mploc_z2x_normalize(z);

	FAST_FREE(buf);
	ASSERT(mploc_z2x_is_good_looking(z));
}

/* prints x to stdout */
void mploc_z2x_print(mploc_z2x_srcptr x)
{
	mp_size_t i,j UNUSED_VARIABLE;
	mpz_t q,r;
	int some=0;

	mpz_init(q);
	mpz_init(r);

	ASSERT(mploc_z2x_is_good_looking(x));
	ASSERT(x->cw!=0 || x->deg==-1);

	for(i=0;i<=x->deg;i++) {
		char c='+';
		MPZ_SET_MPN(q,x->c+i*x->cw,x->cw);
		mploc_limbs_sign_extend(PTR(q),x->cw,x->prec);
		SIZ(q)=x->cw;
		if (PTR(q)[x->cw-1]&(1UL<<(GMP_LIMB_BITS-1))) {
			mploc_limbs_neg(PTR(q),PTR(q),x->cw);
			c='-';
		}
		MPN_NORMALIZE(PTR(q),SIZ(q));
		if (!SIZ(q)) { continue; }
		if (some) printf(" ");
		if (some || c=='-') printf("%c ",c);
		some=1;
#if 0
		printf("(");
		for(j=0;j<x->prec&&(mpz_cmp_ui(q,0)!=0||(j>x->prec-4)); j++) {
			mpz_fdiv_qr_ui(q,r,q,2);
			gmp_printf("%Zd",r);
		}
		if (j<x->prec)
			printf("0<%ld>",x->prec-j);
		printf("...)");
#else
		switch(i) {
			case 0:
				gmp_printf("%Zd",q); break;
			case 1:
				if (mpz_cmp_ui(q,1)!=0) {
					gmp_printf("%Zd*",q);
				}
				printf("X");
				break;
			default:
				if (mpz_cmp_ui(q,1)!=0) {
					gmp_printf("%Zd*",q);
				}
				printf("X^%lu",i);
				break;
		}
#endif
	}

	mpz_clear(q);
	mpz_clear(r);
	if (some) printf(" + ");
	printf("O(2^%d)", (int) x->prec);
}

int mploc_z2x_valuation(mploc_z2x_srcptr x)
{
	mp_size_t aw,i;
	int	v;

	aw=TIGHTCW(x);
	
#ifndef	LOOSE_CW
	ASSERT(aw==x->cw);
#endif

	v=x->prec;
	for(i=0;v && i<=x->deg;i++) {
		v = MIN(mploc_limbs_valuation(x->c+i*x->cw,aw),v);
	}
	return v;
}
		
int mploc_z2x_eq(mploc_z2x_srcptr x, mploc_z2x_srcptr y)
{
	mploc_z2x_t z;
	int r;

	mploc_z2x_init(z,x->coeff_field);
	mploc_z2x_sub(z,x,y);
	r=z->deg==-1;
	mploc_z2x_clear(z);

	return r;
}

int mploc_z2x_eq_ui(mploc_z2x_srcptr x, mp_limb_t v)
{
	mploc_z2x_t z;
	int r;

	if (x->deg)
		return x->deg==-1 && v==0;

	mploc_z2x_init(z,x->coeff_field);
	mploc_z2x_sub_ui(z,x,v);
	r=z->deg==-1;
	mploc_z2x_clear(z);

	return r;
}

/* This does not increase the precision, therefore no realloc is needed
 * (but a normalization step becomes necessary, so the win is tiny).
 */
void mploc_z2x_lshift_quick(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t n)
{
	mp_size_t i;
	mp_size_t wi,wp;

	ASSERT(mploc_z2x_is_good_looking(x));

	if (EXPECT_FALSE(n==0 || x->prec==0)) {
		mploc_z2x_set(z,x);
		return;
	}

	wi = n / GMP_LIMB_BITS;
	wp = n % GMP_LIMB_BITS;

	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,x->deg,x->prec);
	z->deg=x->deg;

	if (EXPECT_FALSE(wp==0)) {
		for(i=0;i<=x->deg;i++) {
			mploc_limbs_move(z->c+i*z->cw+wi,x->c+i*x->cw,x->cw-wi);
			mploc_limbs_zero(z->c+i*z->cw,wi);
		}
	} else {
		for(i=0;i<=x->deg;i++) {
			mpn_lshift(z->c+i*z->cw+wi,x->c+i*x->cw,x->cw-wi,wp);
			mploc_limbs_zero(z->c+i*z->cw,wi);
		}
	}

	mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}
void mploc_z2x_lshift(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t n)
{
	mp_size_t xcw,i;
	mp_size_t wi,wp;
	int needpop;

	ASSERT(mploc_z2x_is_good_looking(x));

	if (EXPECT_FALSE(n==0)) {
		mploc_z2x_set(z,x);
		return;
	} else if (EXPECT_FALSE(x->prec==0)) {
		ASSERT(x->deg==-1);
		mploc_z2x_realloc(z,1,x->deg,n+x->prec);
		z->deg=x->deg;
		/* Since the degree is -1 anyway, we don't have to care
		 * about explicitly padding the coefficients with zeroes
		 */
		return;
	}

	/* Save it now so that self-assignment doesn't kill us */
	xcw = TIGHTCW(x);

	wi = n / GMP_LIMB_BITS;
	wp = n % GMP_LIMB_BITS;

	needpop = (((x->prec-1) % GMP_LIMB_BITS)+1 + wp) > GMP_LIMB_BITS;

	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,x->deg,n+x->prec);
	z->deg=x->deg;

	/* z->cw-wi may exceed x->cw fairly commonly, if x's data once
	 * shifted spans more words than before.
	 */

	ASSERT(TIGHTCW(z)-wi >= xcw && ((TIGHTCW(z)-wi > xcw) == needpop));

	if (EXPECT_FALSE(wp==0)) {
		for(i=0;i<=x->deg;i++) {
			mploc_limbs_move(z->c+i*z->cw+wi,x->c+i*x->cw,xcw);
			mploc_limbs_zero(z->c+i*z->cw,wi);
		}
	} else if (needpop) {
		mp_limb_t pop;
		/* This is somewhat tricky to get right. In case x==z,
		 * then the popping word is after xcw limbs, not after
		 * x->cw limbs... Hence the subtle difference. This
		 * assert is here to triple check that I'm getting it
		 * right.
		 */
		ASSERT(z!=x || TIGHTCW(x)==xcw+(wi+1));
		for(i=0;i<=x->deg;i++) {
			pop=mpn_lshift(z->c+i*z->cw+wi,x->c+i*x->cw,xcw,wp);
			z->c[i*z->cw+wi+xcw]=pop;
			mploc_limbs_zero(z->c+i*z->cw,wi);
		}
	} else {
		/* This branch is perhaps counter-productive. It
		 * embarasses dynamic prediction.
		 * TODO : bench, and probably remove.
		 */
		BAD_CODE_WARNING;
		for(i=0;i<=x->deg;i++) {
			mpn_lshift(z->c+i*z->cw+wi,x->c+i*x->cw,xcw,wp);
			mploc_limbs_zero(z->c+i*z->cw,wi);
		}
	}

	ASSERT(mploc_z2x_is_good_looking(z));
}

void mploc_z2x_rshift(mploc_z2x_ptr z, mploc_z2x_srcptr x, mp_size_t n)
{
	mp_size_t aw,i;
	mp_size_t wi,wp;

	if (EXPECT_FALSE(n==0)) {
		mploc_z2x_set(z,x);
	}

	if (z!=x) mploc_z2x_clobber(z);

	mploc_z2x_realloc(z,1,x->deg,x->prec);
	/* cannot require x->prec-n if z==x :-( */
	z->deg=x->deg;

	aw = TIGHTCW(z);

#ifndef	LOOSE_CW
	ASSERT(aw==z->cw);
#endif

	wi = n / GMP_LIMB_BITS;
	wp = n % GMP_LIMB_BITS;

	if (EXPECT_FALSE(wp==0)) {
		for(i=0;i<=x->deg;i++) {
			mploc_limbs_move(z->c+i*z->cw,x->c+i*x->cw+wi,aw-wi);
		}
	} else {
		for(i=0;i<=x->deg;i++) {
			mpn_rshift(z->c+i*z->cw,x->c+i*x->cw+wi,aw-wi,wp);
		}
	}

	/* NOTE : This needs to be investigated. We don't want to shuffle
	 * cw's for so little.
	 */
	mploc_z2x_realloc(z,-1,z->deg,z->prec-n); /* SHRINK */

	/* Head coeff might have become 0 */
	mploc_z2x_normalize(z);

	ASSERT(mploc_z2x_is_good_looking(z));
}

/****************************************************************************/
/* Multiplication of polynomials */

/* Low-level routines used */

#ifndef TUNING
/* This has been tuned for p3 celeron 700MHz, 128KB cache */
static const mp_size_t kara_mul_thresholds[] = {
	/* cw == 1 */	12,
	/* cw == 2 */	8,
	/* cw == 3 */	6, 6, 6, 6, 6, 6, 6,
	/* cw == 10 */	4, 4, 4, 4, 4};

/* This has _NOT_ been tuned XXX */
static const mp_size_t kara_sqr_thresholds[] = {
	/* cw == 1 */	60,
	/* cw == 2 */	8,
	/* cw == 3 */	6, 6, 6, 6, 6, 6, 6,
	/* cw == 10 */	4, 4, 4, 4, 4};

#define	Z2X_KARA_MUL_THRESHOLD(cw)	(((cw)>=15)?2:kara_mul_thresholds[(cw)-1])
#define	Z2X_KARA_SQR_THRESHOLD(cw)	(((cw)>=15)?2:kara_sqr_thresholds[(cw)-1])

#else

static int kara_mul_threshold=4;
static int kara_sqr_threshold=4;

void z2x_set_kara_mul_threshold(int x)
{
	kara_mul_threshold=x;
}

void z2x_set_kara_sqr_threshold(int x)
{
	kara_sqr_threshold=x;
}

#define Z2X_KARA_MUL_THRESHOLD(cw)	kara_mul_threshold
#define Z2X_KARA_SQR_THRESHOLD(cw)	kara_sqr_threshold

#endif

#define	z2x_mul_n_wrap(r,a,b,cw,d,ws)	((d)<Z2X_KARA_MUL_THRESHOLD(cw))?z2x_mul_basecase(r,a,b,cw,d,d):z2x_kara_mul_n(r,a,b,cw,d,ws)

#define	z2x_sqr_n_wrap(r,a,cw,d,ws)	((d)<Z2X_KARA_SQR_THRESHOLD(cw))?z2x_sqr_basecase(r,a,cw,d):z2x_kara_sqr_n(r,a,cw,d,ws)

#define EXCHPTR(a,na,b,nb)	BEGIN_BLOCK	\
			const mp_limb_t * t;	\
			mp_size_t	n;	\
			t=a;a=b;b=t;		\
			n=na;na=nb;nb=n;	\
		END_BLOCK

/* schoolbook */
static void z2x_mul_basecase(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t na, mp_size_t nb);
static void z2x_sqr_basecase(mp_limb_t * r, const mp_limb_t * a, mp_size_t cw, mp_size_t n);

/* karatsuba */
static void z2x_kara_mul_n(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t n, mp_limb_t * ws);
static void z2x_kara_sqr_n(mp_limb_t * r, const mp_limb_t * a, mp_size_t cw, mp_size_t n, mp_limb_t * ws);

static void z2x_mul_back(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t na, mp_size_t nb);

#if 0
unsigned long long helper(mp_limb_t * x, mp_size_t cw, mp_size_t n)
{
	mpz_t t;
	int i;
	unsigned long long r;

	mpz_init2(t,n);
	mpz_set_ui(t,0);

	for(i=0;i<n;i++) {
		if (*(x+i*cw) & 1)
			mpz_setbit(t,i);
	}
	gmp_printf("  0x%Zx  ",t);
	r = mpz_get_ui(t);
	mpz_fdiv_q_2exp(t,t,32);
	r = r | ((unsigned long long)(mpz_get_ui(t))<<32ULL);
	mpz_clear(t);
	return r;
}

unsigned long long helper_p(mploc_z2x_srcptr x)
{
	return helper(x->c,x->cw,x->deg+1);
}
#endif

/* kara_square ; wins 2x if basecase manages to win that much. */
void mploc_z2x_sqr(mploc_z2x_ptr z, mploc_z2x_srcptr x)
{
	mploc_z2x_t tx;
	mp_limb_t * tbuf;

	BAD_CODE_WARNING;
	/* thresholds have *NOT* been optimized !!! */

	if (x->deg==-1) {
		z->deg=-1;
		mploc_z2x_realloc(z,1,-1,x->prec+1);
		return;
	}

	mploc_z2x_init(tx,x->coeff_field);
	mploc_z2x_set_change_precision(tx,x,x->prec+1);

	if (z!=x) mploc_z2x_clobber(z);
	mploc_z2x_realloc(z,1,2*tx->deg,tx->prec);

	mploc_z2x_realloc(z,-1,2*tx->deg,tx->prec); /* SHRINK */

	ASSERT(z->cw==tx->cw);

	tbuf = malloc((tx->deg+1) * tx->cw * sizeof(mp_limb_t));

	z2x_sqr_n_wrap(z->c,tx->c,tx->cw,tx->deg+1,tbuf);

	z->deg = 2 * tx->deg;

	free(tbuf);

	mploc_z2x_clear(tx);

	mploc_z2x_normalize(z);
}

/* This is the front-end. One must first tweak the input so that
 * everything is reasonably uniform (notably the cw's)
 *
 * XXX This function does a _shrink() ; Perhaps it could be better to
 * remove it.
 *
 * The reason for the shrink is that the _kara routines rely on cw being
 * the same all along the way. So since we have to align to some value at
 * some point, it's better to force the smallest possible value.
 */
void mploc_z2x_mul(mploc_z2x_ptr z, mploc_z2x_srcptr x, mploc_z2x_srcptr y)
{
	mploc_z2x_ptr tmp_x=NULL, tmp_y=NULL;
	mp_limb_t * target = NULL;
	mp_size_t totaldeg;

	/* This is a temporary hack, in order to discourage the user from
	 * using this function knowingly as a squaring function
	 */
	if (x==y && z!=x) {
		fprintf(stderr,"Warning: doing a square\n");
	}

	totaldeg=(x->deg==-1||y->deg==-1)?-1:(x->deg + y->deg);

	if (totaldeg==-1) {
		/* Do a ``soft clobber'', since we don't want to thrash
		 * the precision */
		z->deg=-1;
		/* also set the precision */
		mploc_z2x_realloc(z,1,-1,MIN(x->prec,y->prec));
#ifndef	LOOSE_CW
		mploc_z2x_realloc(z,-1,z->deg,z->prec); /* SHRINK */
#endif
		return;
	}

	/* If z is one of the operands, this saves some work. */
	if (z!=x && z!=y) mploc_z2x_clobber(z);

	mploc_z2x_realloc(z,1,totaldeg,MIN(x->prec,y->prec));
	/* Don't set z->deg now... */

	mploc_z2x_realloc(z,-1,totaldeg,MIN(x->prec,y->prec)); /* SHRINK */
	
	/* rely on the generic stuff ; playing with memory structure is
	 * pointless here (win 1e-5% ? hmmm...)
	 */
	if (z->cw != x->cw) {
		tmp_x = malloc(sizeof(mploc_z2x_t));
		mploc_z2x_init(tmp_x,x->coeff_field);
		mploc_z2x_set_change_precision(tmp_x,x,z->prec);
	}

	if (z->cw != y->cw) {
		tmp_y = malloc(sizeof(mploc_z2x_t));
		mploc_z2x_init(tmp_y,y->coeff_field);
		mploc_z2x_set_change_precision(tmp_y,y,z->prec);
	}

	if ((!tmp_x && z->c == x->c) || (!tmp_y && z->c == y->c)) {
		target = malloc(z->cw * (totaldeg+1) * sizeof(mp_limb_t));
	}


	/* time to call the back end */
	z2x_mul_back(target?target:z->c,(tmp_x?tmp_x:x)->c,(tmp_y?tmp_y:y)->c,
			z->cw,x->deg+1,y->deg+1);

	/* NOW we can set z->deg */
	z->deg = totaldeg;

	
	if (tmp_x) { mploc_z2x_clear(tmp_x); free(tmp_x); }
	if (tmp_y) { mploc_z2x_clear(tmp_y); free(tmp_y); }
	if (target) {
		mploc_limbs_copy(z->c,target,z->cw * (z->deg + 1));
		free(target);
	}

	mploc_z2x_normalize(z);
}

/* Helpers for add and sub
 *
 * TODO: could actually be useful for mploc_z2x_add too.
 */
static inline void z2x_add(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t n)
{
	mp_size_t i;
	for(i=0;i<n;i++) {
		mpn_add_n(r,a,b,cw);
		r+=cw;
		a+=cw;
		b+=cw;
	}
}

static inline void z2x_sub(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t n)
{
	mp_size_t i;
	for(i=0;i<n;i++) {
		mpn_sub_n(r,a,b,cw);
		r+=cw;
		a+=cw;
		b+=cw;
	}
}

static inline void z2x_neg(mp_limb_t * r, const mp_limb_t * a, mp_size_t cw, mp_size_t n)
{
	mp_size_t i;
	for(i=0;i<n;i++) {
		mploc_limbs_neg(r,a,cw);
		r+=cw;
		a+=cw;
	}
}

/*
 * This is the code for unbalanced mults. It does its best to to the good
 * number of symmetric multiplications.
 *
 * This follows mpn_mul from gmp-4.1.
 *
 * na and nb here are the number of coefficients involved (hence degree+1).
 *
 * na and nb multiplied form na+nb-1 coefficients (pain...)
 */
static void z2x_mul_back(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t na, mp_size_t nb)
{
	mp_limb_t * tmp, * kara_tmp;
	mp_size_t l;

	if (nb > na)	EXCHPTR(a,na,b,nb);

	if (nb<Z2X_KARA_MUL_THRESHOLD(cw)) {
		z2x_mul_basecase(r,a,b,cw,na,nb);
		return;
	}

	/*
	 * We're going to do Karatsuba with size at most n. Let's
	 * allocate the temp. buffer.
	 */

	kara_tmp = malloc((nb) * cw * sizeof(mp_limb_t));

	z2x_kara_mul_n(r,a,b,cw,nb, kara_tmp);

	if (na!=nb) {
		r  += nb*cw;
		na -= nb;
		a  += nb*cw;
		l   = nb-1;

		/* l is the number of coefficients set in r by the last
		 * mult.  operation. Of course, since r might have been
		 * shifted, l takes this into account. Since we're
		 * swapping pointers back and forth, there's not much
		 * point in trying to ``guess'' what l is from the
		 * current values of na and nb (and it would be pretty
		 * hard, in fact).
		 */
		
		if (na < nb) EXCHPTR(a,na,b,nb);

		/* XXX tmp and kara_tmp should be merged. */
		tmp = FAST_ALLOC(((nb>=Z2X_KARA_MUL_THRESHOLD(cw)?nb:na)+nb-1)
				*cw*sizeof(mp_limb_t));
		
		while (nb>=Z2X_KARA_MUL_THRESHOLD(cw)) {
			
			/* Loop invariant here: the product of a and b,
			 * with respectively na and nb coefficients, has
			 * to be stored into r, where l coefficents are
			 * already set.
			 */

			z2x_kara_mul_n(tmp,a,b,cw,nb, kara_tmp);

			if (l<= (2*nb - 1)) {
				z2x_add(r,r,tmp,cw,l);
				memcpy(r+l*cw,tmp+l*cw,
					(2*nb-1-l)*cw*sizeof(mp_limb_t));
				l=2*nb-1;
			} else {
				z2x_add(r,r,tmp,cw,2*nb-1);
			}

			r  += nb*cw;
			a  += nb*cw;
			na -= nb;
			l  -= nb;
			
			if (na < nb) EXCHPTR(a,na,b,nb);
		}
		
		if (nb != 0) {
			z2x_mul_basecase(tmp,a,b,cw,na,nb);
			if (l <= na + nb - 1) {
				z2x_add(r,r,tmp,cw,l);
				memcpy(r+l*cw,tmp+l*cw,
				(na + nb - 1 - l)*cw*sizeof(mp_limb_t));
			} else {
				z2x_add(r,r,tmp,cw,na + nb-1);
			}
		}

		FAST_FREE(tmp);
	}

	free(kara_tmp);
}

/* the number of bits required to store x */
static int ilog2(mp_size_t x) {
  int l;
  l = 1;
  x >>= 1;
  while (x) {
    l++;
    x >>= 1;
  }
  return l;
}

/* Pack and Pad (a.k.a. Kronecker substitution)
 * written by PG, could be used some day for large operands
 */
void mploc_z2x_packpad_mul(mploc_z2x_ptr z, mploc_z2x_srcptr x, mploc_z2x_srcptr y)
{
        mploc_z2x_t tmp_x, tmp_y;
        mp_size_t mindeg, totaldeg;
        mp_size_t prec, padded_prec;
        mp_size_t cw;

        mploc_z2x_init(tmp_x, x->coeff_field);
        mploc_z2x_init(tmp_y, x->coeff_field);

        prec = MIN(x->prec, y->prec);
        mindeg = MIN(x->deg, y->deg);
        totaldeg = x->deg+y->deg;
        ASSERT (mindeg > 1  &&  prec >= 1);
        padded_prec = 2*prec + ilog2(mindeg+1);

        /* pack-pad */
        mploc_z2x_set_change_precision(tmp_x, x, padded_prec);
        mploc_z2x_insert_zero(tmp_x, prec);
        mploc_z2x_set_change_precision(tmp_y, y, padded_prec);
        mploc_z2x_insert_zero(tmp_y, prec);

        mploc_z2x_realloc(z,1,totaldeg,padded_prec);
        mploc_z2x_realloc(z,-1,totaldeg,padded_prec);
        cw = z->cw; 
        ASSERT (z->cw == tmp_x->cw && z->cw == tmp_y->cw);

        /* mul */
        mpn_mul(z->c, tmp_x->c, (tmp_x->deg+1)*cw,
            tmp_y->c, (tmp_y->deg+1)*cw);

	/* unpack */
        z->deg = totaldeg;
        mploc_z2x_normalize(z);
        mploc_z2x_force_precision(z, prec);

        mploc_z2x_clear(tmp_x);
        mploc_z2x_clear(tmp_y);
}


/* karatsuba ; multiply two polynomials holding n coefficients , produce
 * one having 2n-1 coefficients.
 */
static void z2x_kara_mul_n(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t n, mp_limb_t * ws)
{
	/* ws must hold room for n * cw words if n is odd, and only
	 * (n-1) * cw words if n is even
	 */

	if (n&1)	{	// odd number of coeffs.
		mp_size_t p, of;

		p = n>>1;
		/* q = p + 1 */

		of = p * cw;

		/* A = al + X^p * ah    deg al = p-1, deg ah = q-1
		 * B = bl + X^p * bh    deg bl = p-1, deg bh = q-1
		 *
		 * alpha = al - ah      deg alpha = q-1
		 * beta  = bh - bl      deg beta =  q-1
		 *
		 * gamma = alpha * beta + al*bl + ah * bh = al * bh + ah * bl
		 *
		 * A*B = al * bl + X^p * gamma + X^2p * ah * bh
		 */
		z2x_sub(r,a,a+of,cw,p);
		mploc_limbs_neg(r+of,a+(n-1)*cw,cw);

		/* Now r[0...q[ is alpha = al-ah */

		z2x_sub(r+of+cw,b+of,b,cw,p);
		mploc_limbs_copy(r+n*cw,b+(n-1)*cw,cw);// p+q=n XXX

		/* Now r[q...2q[ is beta = bh - bl */
		/* Up to n + 1 coefficients in r are used. */

		z2x_mul_n_wrap(ws,r,r+of+cw,cw,p+1,r+(n+1)*cw);
		z2x_mul_n_wrap(r+2*of,a+of,b+of,cw,p+1,r);

		z2x_add(ws,ws,r+2*of,cw,n);
		memcpy(r+of,ws,of*sizeof(mp_limb_t));
		z2x_add(r+2*of,r+2*of,ws+of,cw,p+1);

		/* Now r[p...2n-1[ contains
		 * X^p*(alpha * beta + ah*bh) + X^2p*ah*bh
		 */

		z2x_mul_n_wrap(ws,a,b,cw,p,r);

		memcpy(r,ws,of*sizeof(mp_limb_t));
		z2x_add(r+of,r+of,ws+of,cw,p-1);
		z2x_add(r+of,r+of,ws,cw,n-2);
	} else {	// Even n
		mp_size_t p, of;

		p = n>>1;	// and q = p

		of = p * cw;

		z2x_sub(r,a,a+of,cw,p);
		z2x_sub(r+of,b+of,b,cw,p);

		z2x_mul_n_wrap(ws,r,r+of,cw,p,r+n*cw);
		z2x_mul_n_wrap(r+2*of,a+of,b+of,cw,p,r);

		z2x_add(ws,ws,r+2*of,cw,n-1);
		memcpy(r+of,ws,of*sizeof(mp_limb_t));
		z2x_add(r+2*of,r+2*of,ws+of,cw,p-1);

		z2x_mul_n_wrap(ws,a,b,cw,p,r);

		memcpy(r,ws,of*sizeof(mp_limb_t));
		z2x_add(r+of,r+of,ws+of,cw,p-1);
		z2x_add(r+of,r+of,ws,cw,n-1);
#if 0
		/* This is completely generic */
	} else {
		mp_size_t p, q, of;
		p = n>>1;
		q = n - p;
		of = p * cw;
		z2x_sub(r,a,a+of,cw,p);
		if (q!=p) mploc_limbs_neg(r+of,a+(n-1)*cw,cw);
		z2x_sub(r+q*cw,b+of,b,cw,p);
		if (q!=p) mploc_limbs_copy(r+n*cw,b+(n-1)*cw,cw);// p+q=n XXX
		z2x_mul_n_wrap(ws,r,r+q*cw,cw,q,r+2*q*cw);
		z2x_mul_n_wrap(r+2*of,a+of,b+of,cw,q,r);
		z2x_add(ws,ws,r+2*of,cw,2*q-1);
		memcpy(r+of,ws,of*sizeof(mp_limb_t));
		z2x_add(r+2*of,r+2*of,ws+of,cw,2*q-1-p);
		z2x_mul_n_wrap(ws,a,b,cw,p,r);
		memcpy(r,ws,of*sizeof(mp_limb_t));
		z2x_add(r+of,r+of,ws+of,cw,p-1);
		z2x_add(r+of,r+of,ws,cw,2*p-1);
#endif
	}
}

#define	z2x_mul2_n_wrap(r,a,b,cw,da,flip,ws)	((da)<Z2X_KARA_MUL_THRESHOLD(cw))?z2x_mul_basecase(r,a,b,cw,da,(da)-(flip)):z2x_kara2_mul_n(r,a,b,cw,da,flip,ws)

/* karatsuba2 ; multiply two polynomials holding respectively n and
 * (n-flip) coefficients , produce one having 2n-flip-1 coefficients.
 *
 * THIS ROUTINE IS __**NOT**__ GENERIC: IT ASSUMES THAT flip=0 or 1 !!!
 *
 * (it's a useful tool to have for the division)
 */
static void z2x_kara2_mul_n(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t n, mp_size_t flip, mp_limb_t * ws)
{
	mp_size_t s, t, of;
	mp_size_t odd = n & 1UL;

	s = n >> 1UL;
	t = n - s;
	of = s * cw;

	z2x_sub(r,a,a+of,cw,s);
	if (odd) mploc_limbs_neg(r+of,a+(n-1)*cw,cw);

	z2x_sub(r+t*cw,b+of,b,cw,t-(flip|odd));
	/* We're really copying c = MIN(s,n-flip-s) == MIN(s,t-flip) coeffs
	 * odd == 0, flip == 0 ----> c=s=t-flip, ok.	(c=s=t)
	 * odd == 0, flip == 1 ----> c=t-1 --> 1 neg	(c=s-1=t-1)
	 * odd == 1, flip == 0 ----> c=s --> 1 copy	(c=s=t-1)
	 * odd == 1, flip == 1 ----> c=s=t-flip, ok	(c=s=t-1)
	 *
	 * Note that c rewrites as t-(flip|odd)
	 *
	 * The _new_ c value is, respectively for each case:
	 * 	s   = t
	 * 	s   = t
	 * 	s+1 = t
	 * 	s   = t-1
	 *
	 * Hence the new flip value for the cross product is flip & odd
	 */
	if (~odd & flip) mploc_limbs_neg(r+(n-1)*cw,b+(s-1)*cw,cw);
	if (odd & ~flip) mploc_limbs_copy(r+n*cw,b+(n-1)*cw,cw);
	/* busy : r[0..t[ and r[t..2t-(flip&odd)[ */
	z2x_mul2_n_wrap(ws,r,r+t*cw,cw,t,flip & odd, r+(2*t-(flip&odd))*cw);
	/* busy : ws[0..2*t-(flip&odd)-1[ ; contains X product */
	/* put high part in right place : r[2s..2n-1-flip[ */
	z2x_mul2_n_wrap(r+2*of,a+of,b+of,cw,t,flip,r);
	/* r[2s..[ has 2t-1-flip coeffs (ah * bh) */
	/* w[..[ has 2t-1-(flip&odd) coeffs, which is >= 2t-1-flip */
	z2x_add(ws,ws,r+2*of,cw,2*t-1-flip);
	/* ws[0..2t-1-(flip&odd)[ = alpha * beta + ah * bh */

	mploc_limbs_copy(r+of,ws,of);
	z2x_add(r+2*of,r+2*of,ws+of,cw,2*t-1-(flip&odd)-s);
	/* busy: r[s..s+2t-1-(f&o)[ == alpha * beta + ah * bh + X^s * ah * bh */

	z2x_mul2_n_wrap(ws,a,b,cw,s,0,r);
	mploc_limbs_copy(r,ws,of);
	z2x_add(r+of,r+of,ws+of,cw,s-1);
	z2x_add(r+of,r+of,ws,cw,2*s-1);
}

/* karatsuba squaring */
static void z2x_kara_sqr_n(mp_limb_t * r, const mp_limb_t * a, mp_size_t cw, mp_size_t n, mp_limb_t * ws)
{
	mp_size_t p, q, of;

	p = n>>1;
	q = n - p;
	of = p * cw;

	z2x_sub(r,a+of,a,cw,p);
	if (q!=p) mploc_limbs_copy(r+of,a+(n-1)*cw,cw);
	
	z2x_sqr_n_wrap(ws,r,cw,q,r+2*q*cw);
	z2x_neg(ws,ws,cw,2*q-1);

	z2x_sqr_n_wrap(r+2*of,a+of,cw,q,r);
	z2x_add(ws,ws,r+2*of,cw,2*q-1);
	memcpy(r+of,ws,of*sizeof(mp_limb_t));
	z2x_add(r+2*of,r+2*of,ws+of,cw,2*q-1-p);
	z2x_sqr_n_wrap(ws,a,cw,p,r);
	memcpy(r,ws,of*sizeof(mp_limb_t));
	z2x_add(r+of,r+of,ws+of,cw,p-1);
	z2x_add(r+of,r+of,ws,cw,2*p-1);
}

/* schoolbook */
static void z2x_mul_basecase(mp_limb_t * r, const mp_limb_t * a, const mp_limb_t * b, mp_size_t cw, mp_size_t na, mp_size_t nb)
{
	mp_size_t i,j;
	mp_limb_t * s;
	s = FAST_ALLOC(3*cw*sizeof(mp_limb_t));

	ASSERT(r!=a && r!=b);

	ASSERT(na && nb);

	memset(r,0,(na+nb-1)*cw*sizeof(mp_limb_t));
#define	MULCODE(cw)				\
	for(i=0;i<na;i++) {			\
		mp_limb_t * cr;			\
		const mp_limb_t * cb;		\
		cr = r + i*cw;			\
		cb = b;				\
		for(j=0;j<nb;j++) {		\
			MUL ## cw(s,a,cb);	\
			ADD ## cw(cr,s);	\
			cr+=cw;			\
			cb+=cw;			\
		}				\
		a+=cw;				\
	}
#define	ADD1(t,s) t[0]+=s[0]
#define	ADD2(t,s) mpn_add_n(t,t,s,2)
#define	ADD3(t,s) mpn_add_n(t,t,s,3)
#define	ADD4(t,s) mpn_add_n(t,t,s,4)
#define	ADD5(t,s) mpn_add_n(t,t,s,5)
#define	ADD6(t,s) mpn_add_n(t,t,s,6)
#define	ADD7(t,s) mpn_add_n(t,t,s,7)
#define	ADD8(t,s) mpn_add_n(t,t,s,8)
	switch(cw) {
		case 1: MULCODE(1); break;
		case 2: MULCODE(2); break;
		case 3: MULCODE(3); break;
		case 4: MULCODE(4); break;
		case 5: MULCODE(5); break;
		case 6: MULCODE(6); break;
		case 7: MULCODE(7); break;
		case 8: MULCODE(8); break;
		default:
			for(i=0;i<na;i++) {
				mp_limb_t * cr;
				const mp_limb_t * cb;
				cr = r + i*cw;
				cb = b;
				for(j=0;j<nb;j++) {
					mploc_limbs_short_product(
							s,a,cb,s+cw,cw);
					mpn_add_n(cr,cr,s,cw);
					cr+=cw;
					cb+=cw;
				}
				a+=cw;
			}
	}

	FAST_FREE(s);
}

static void z2x_sqr_basecase(mp_limb_t * r, const mp_limb_t * a, mp_size_t cw, mp_size_t n)
{
	mp_size_t i,j;
	mp_limb_t * s;
	s = FAST_ALLOC(3*cw*sizeof(mp_limb_t));

	ASSERT(r!=a);
	ASSERT(n);

	memset(r,0,(2*n-1)*cw*sizeof(mp_limb_t));

	switch(cw) {
		case 1:
			for(i=0;i<n;i++) for(j=0;j<i;j++) {
				r[i+j]+=a[i]*a[j];
			}

			for(i=0;i<2*n-1;i++) {
				r[i]<<=1;
			}

			for(i=0;i<n;i++) {
				r[0]+=a[0]*a[0];
				a++;
				r++;
				r++;
			}
			break;
		case 2:
			/* 20040825 ; code in this branch was badly
			 * broken ! I overlooked the fact that of course,
			 * long shifts and squares were an absolute
			 * necessity.
			 */
			for(i=0;i<n;i++) for(j=0;j<i;j++) {
				ADDMUL2(r+(i+j)*2,a+i*2,a+j*2);
			}

			for(i=0;i<2*n-1;i++) {
				mpn_lshift(r+(i<<1),r+(i<<1),2,1);
			}

			for(i=0;i<n;i++) {
				SQR2(s,a);
				mpn_add_n(r,r,s,2);
				a+=2;
				r+=4;
			}
			break;
		default:
			for(i=0;i<n;i++) for(j=0;j<i;j++) {
				mploc_limbs_short_product(
						s,a+i*cw,a+j*cw,
						s+cw,cw);
				mpn_add_n(r+(i+j)*cw,r+(i+j)*cw,s,cw);
			}

			for(i=0;i<2*n-1;i++) {
				mpn_lshift(r+i*cw,r+i*cw,cw,1);
			}

			for(i=0;i<n;i++) {
				mploc_limbs_short_square(s,a+i*cw,s+cw,cw);
				mpn_add_n(r+2*i*cw,r+2*i*cw,s,cw);
			}

	}

	FAST_FREE(s);
}



/* u <- [X^i] x */
/* This variant is used in mploc_z2x_div. This is useful when 
 *   b->prec   is much larger   a->prec
 * which happens quite often in AGM-like algorithms
 */
static void mploc_z2x_get_coeff_change_precision(mploc_ptr a,
    mploc_z2x_srcptr x, mp_size_t i, mp_size_t k)
{
	ASSERT(i>=0);
	ASSERT(k<= x->prec);

	mploc_z2_force_precision(a, k);
	if (i<=x->deg)
		mploc_limbs_copy(PTR(a->x),x->c+i*x->cw,SIZ(a->x));
	else
		mploc_limbs_zero(PTR(a->x),SIZ(a->x));

	ASSERT(mploc_z2_is_good_looking(a));
}

int mploc_z2x_coeff_eq_ui(mploc_z2x_srcptr x, mp_size_t i, mp_limb_t v)
{
	ASSERT(mploc_z2x_is_good_looking(x));
	
	ASSERT(i>=0);
	if (EXPECT_FALSE(i>=x->deg))
		return v==0;

	return x->c[i*x->cw]==v &&
		mploc_limbs_valuation(x->c+i*x->cw+1,x->cw-1) >= x->prec - GMP_LIMB_BITS;
}

/* Compute the inverse of the formal power series represented by src ; we
 * start with the easy approximation at degree 1, and then climb up with
 * newton iteration.
 * Note that it is not possible to start with a run of the Euclidean
 * algorithm, for an excellent reason : z2 is not a field.
 */
void mploc_z2x_series_inv(mploc_z2x_ptr dst,
		mploc_z2x_srcptr src,
		mp_size_t deg)
{
	mploc_z2x_t half;
	mploc_z2x_t y,t;

	if (!mploc_z2x_coeff_eq_ui(src,0,1))
		DIVIDE_BY_ZERO();
	
	if (deg <= 1) { 
		mploc_t c;
		mploc_z2_init(c,src->coeff_field);
		mploc_z2x_get_coeff(c,src,0);
		/* This will divide by zero if c is not invertible */
		mploc_z2_inv(c,c);
		mploc_z2x_set_coeff(dst,0,c);
		mploc_z2_clear(c);
		return;
	}

	mploc_z2x_init(half,src->coeff_field);
	mploc_z2x_init(y,src->coeff_field);
	mploc_z2x_init(t,src->coeff_field);

	mploc_z2x_realloc(half,1,deg,src->prec);

	mploc_z2x_series_inv(half,src,deg - deg/2);
	mploc_z2x_set(t,src);

	mploc_z2x_chop_deg(t,deg);
	mploc_z2x_chop_deg(half,deg);

	mploc_z2x_mul(y,t,half);
	mploc_z2x_sub_ui(y,y,1);
	mploc_z2x_mul(y,y,half);
	mploc_z2x_chop_deg(y,deg);

	mploc_z2x_sub(dst,half,y);

	mploc_z2x_clear(half);
	mploc_z2x_clear(y);
	mploc_z2x_clear(t);
}

/* Compute the ``preinverse'' of a polynomial, defined as the inverse in
 * R[[X]] of the reciprocal of f.
 *
 * The preinverse Q is such that Q * \hat f = 1 + O(X^\deg f + 1)
 */
void mploc_z2x_preinv(mploc_z2x_ptr dst, mploc_z2x_srcptr src)
{
	mploc_z2x_t tmp;
	mploc_z2x_init(tmp,src->coeff_field);
	mploc_z2x_reciprocal(tmp,src);
	mploc_z2x_series_inv(dst,tmp,src->deg);
	mploc_z2x_clear(tmp);
}
#define	MUL(r,a,b,t,n)	mploc_limbs_short_product(r,a,b,t,n)

/* q or r might be unneeded. */
void mploc_z2x_div(mploc_z2x_ptr q, mploc_z2x_ptr r,
		mploc_z2x_srcptr a, mploc_z2x_srcptr b)
{
	mploc_z2x_t	ta;
	mp_size_t	np,cw;
	mploc_t		c;
	mp_limb_t     * buf;	/* For multiplications */
	mp_size_t	i,j;

	if (b->deg==-1)	DIVIDE_BY_ZERO();

	np=MIN(a->prec,b->prec);

	ASSERT(q || r);	/* Otherwise user is an idiot */
			/* XXX Do remove if user *is* an idiot :-) */


	ASSERT(q!=b && r!= b);

	if (q) {
		if (q!=a && q!=b) mploc_z2x_clobber(q);
		mploc_z2x_realloc(q,1,MAX(a->deg-b->deg,q->deg),np);
	}
	if (r) {
		if (r!=a && r!=b) mploc_z2x_clobber(r);
		mploc_z2x_realloc(r,1,MAX(r->deg,b->deg-1),np);
	}
	if (q && r)
		ASSERT(q!=r);

	if (a->deg < b->deg) {
		/* Set r before, in case q==a ! */
		if (r) mploc_z2x_set_change_precision(r,a,np);
		if (q) q->deg=-1;
		if (q) ASSERT(mploc_z2x_is_good_looking(q));
		if (r) ASSERT(mploc_z2x_is_good_looking(r));
		return;
	}

	mploc_z2x_init(ta,a->coeff_field);
	mploc_z2x_set_change_precision(ta,a,np);

	mploc_z2_init(c,a->coeff_field);

	mploc_z2x_get_coeff_change_precision(c,b,b->deg,np);
	mploc_z2_inv(c,c);

	cw = ta->cw;	/* New width */

	ASSERT(SIZ(c->x)==cw);
	ASSERT(b->cw>=cw);

	buf = FAST_ALLOC(4 * cw * sizeof(mp_limb_t));

	if (q) q->deg=a->deg-b->deg;
	if (r) r->deg=b->deg-1;

	if (q) ASSERT(q->cw >= cw);

	if (q) mploc_limbs_insert_zero(q->c,(q->deg+1)*q->cw,0);
	if (r) mploc_limbs_insert_zero(r->c,(r->deg+1)*r->cw,0);

	/* i is merely an alias fo ta->deg here XXX change this to a
	 * macro */
	for(i=ta->deg;i>=b->deg;i=ta->deg) {
		// ASSERT(i==ta->deg);
		MUL(buf+2*cw,ta->c+i*cw,PTR(c->x),buf,cw);
		for(j=1;j<=b->deg;j++) {
			MUL(buf+3*cw,buf+2*cw,b->c+(b->deg-j)*b->cw,buf,cw);
			mpn_sub_n(ta->c+(i-j)*cw,ta->c+(i-j)*cw,
					buf+3*cw,cw);
		}
		ta->deg--;	/* Drop head coeff */
		mploc_z2x_normalize(ta);
		if (q) mploc_limbs_copy(q->c+(i-b->deg)*q->cw,buf+2*cw,cw);
	}
	if (r) for(j=0;j<b->deg;j++) {
		mploc_limbs_copy(r->c+j*r->cw,ta->c+j*cw,cw);
	}
	/* No need to normalize q */
	if (r) mploc_z2x_normalize(r);

	FAST_FREE(buf);
	mploc_z2_clear(c);
	mploc_z2x_clear(ta);

	if (q) ASSERT(mploc_z2x_is_good_looking(q));
	if (r) ASSERT(mploc_z2x_is_good_looking(r));
}

/* Recursive division.
 *
 * Input : deg(b) = n ; deg(a) <= n + na - 1
 * Constraint : na <= n ;
 *
 * over-commented, of course, but it's so messy that I'm frightened with
 * getting it wrong at basically each and every place in the file.
 *
 */
/* This version assumes that arguments do not overlap, and can hold the
 * following space:
 *
 * a[0..2n[ ; significant stuff [0..n+na[
 *          --> note that a[2n-1] has to be zero if na==n-1
 * b[0..n+1[
 * q[0..n[ ; this is required. q[0..na[ is not enough.
 * r[0..n[
 *
 * ws[0..F(n,na)[
 */
void z2x_div_rec_inner(mp_limb_t * q, mp_limb_t * r,
		mp_limb_t * a, mp_limb_t * b,
		mp_size_t cw,
		mp_size_t n,
		mp_size_t na,
		mp_limb_t * ws)
{
	mp_size_t s = n/2;
	mp_size_t t = n - s;
	

	/*
	 * Corresponding Magma code:
	 *
	 * bh,bl:=Quotrem(b,X^s);
	 * ah,al:=Quotrem(a,X^(2*s));
	 * qh,__rh:=recdivrem(ah,bh);
	 * z:=__rh * X^s - bl * qh;
	 * ql,__tail:=recdivrem(z,bh);
	 * r:=__tail*X^s+al-ql*bl;
	 * return qh*X^s+ql,r;
	 */
	/* split rules, number of coeffs for hi/lo parts:
	 * a = ah * X^(2*s) + al	<= 2t	2s
	 * b = bh * X^s + bl		t+1	s
	 * q = qh * X^s + ql		t	s
	 */

#define	ah	(a+2*s*cw)
#define	al	(a)
#define	bh	(b+s*cw)
#define	bl	(b)
#define	qh	(q+s*cw)
#define	ql	(q)

#define	K	(t)
	/* First, compute the high division:
	 *
	 * ah = qh * bh + __rh
	 *
	 * Number of coefficients of a and b known from ah and bh :
	 * size(ah) = 2t
	 * size(bh) = t + 1
	 *
	 * This is tuned to fit in the rule degree(2n-1) // degree(n)
	 * division.
	 *
	 * Since size(bh) >= size(qh) , the computed quotient _really_ is qh.
	 */
	 
	/* size(__rh) = t */
	
	/* Let K be F(t,t)
	 *
	 * K >= F(t,na-s)
	 * K >= F(t,s)
	 *
	 */
	/* Uses t in ws */
	z2x_div_rec_inner(qh,r + s * cw,ah,bh,cw,t,na-s,ws);
	/* qh in q[s..na[
	 * __rh in r[s..n[
	 */
	mploc_limbs_zero(q+na*cw,(n-na)*cw);

	/* The next recursive call will compute ql and __tail ; __tail, in
	 * turn, will be used to infer the remainder r.
	 *
	 * We have      __head = bh * ql + __tail
	 * where:
	 *                  __head = __rh X^s - bl * qh
	 *                  __tail = (bl * ql + r - al) * X^-s
	 *
	 * and r is the remainder we're looking for.
	 *
	 * __head has size MAX(s + t, s + na-s-1) == s + t ; since we
	 * chose s = floor, this is <= 2t Note that it is the first
	 * occurence of the property s == floor.
	 * 
	 * __tail has size t (easily checked)
	 *
	 * The remainder comes out simply with:
	 *
	 * 	r = __tail * X^s + al - bl * ql
	 *
	 * size(bh) = t+1
	 * size(bl) = s
	 * size(qh) = na-s <= t
	 * size(ql) = s
	 * size(__rh) = t
	 * size(r) = n
	 *
	 * Note that bl*qh is a bit unbalanced : s * t ; there may be a
	 * difference of 1.
	 */

	/* bl*qh has size s+t-1 = n-1 ; hence it fits inside r[0..n[ */
	/* __rh is in r[s..n[ for now ; save r[s..2s[ */
	mploc_limbs_copy(ws+K*cw,r+s*cw,s*cw);
	mploc_limbs_zero(r,2*s*cw);
	/* __rh in ws[K..K+s[ and r[2s..n[ (second area may be null) */

	/* ASSUMES that K >= t ; uses t in ws */
	z2x_mul2_n_wrap(r,qh,bl,t,t-s,cw,ws);
	z2x_neg(r,r,cw,n-1);
	
	/* add __rh ; the high coeff is already in place if it exists */
	z2x_add(r+s*cw,r+s*cw,ws+K*cw,cw,s);

	/* ws is now totally clean ; __head is in r[0..n[ */

	/* Uses s + K in ws ; quotient fits in ql. */
	z2x_div_rec_inner(ql,ws+K*cw,r+(t-s)*cw,bh+(t-s)*cw,cw,s,s,ws);
	mploc_limbs_copy(r+(n-1)*cw,r,(t-s)*cw);

	/*
	 * Uses t in ws ; assumes K >= t
	 */
	z2x_mul2_n_wrap(r,bl,ql,t,t-s,cw,ws);
	z2x_neg(r,r,cw,n-1);

	mploc_limbs_copy(ws+(K-1)*cw,r+(n-1)*cw,(t-s)*cw);
	mploc_limbs_zero(r+(n-1)*cw,cw);

	z2x_add(r+s*cw,r+s*cw,ws+K*cw,cw,t);
	z2x_add(r,r,a,cw,2*s);

	/* Hence F must satisfy:
	 *
	 * F(n,na) >= t + F(s, s)
	 *
	 * Chosing F(n,*) = n fits.
	 */
#undef	ah
#undef	al
#undef	bh
#undef	bl
#undef	qh
#undef	ql
}
